import os
import sys
import contextlib


@contextlib.contextmanager
def mute_stdout_stderr():
    """Context manager to redirect stdout and stderr to devnull. Only used to prevent
    terminal flooding in unittests. If an error is raised and not caught before, this
    will hard-exit out"""

    fail = False
    outnull = open(os.devnull, "w")
    old_stdout = sys.stdout
    old_stderr = sys.stderr
    sys.stdout = outnull
    sys.stderr = outnull
    try:
        yield
    except Exception:  # Bare exception to exit without printing anything
        fail = True
    finally:
        sys.stdout = old_stdout
        sys.stderr = old_stderr
        outnull.close()
        if fail:
            sys.exit(2)



from packaging import version
import pandas as pd
import numpy as np
import numpy.typing as npt
import logging
import scipy.ndimage as ndimage
import pathlib
from pytom_tm.tmjob import TMJob
from pytom_tm.mask import spherical_mask
from pytom_tm.angles import get_angle_list, convert_euler
from pytom_tm.dataclass import RelionTiltSeriesMetaData
from pytom_tm.io import read_mrc
from scipy.special import erfcinv
from scipy.optimize import curve_fit
from tqdm import tqdm


plotting_available = False
try:
    import matplotlib.pyplot as plt
    import seaborn as sns

    sns.set(context="talk", style="ticks")
    plotting_available = True
except ModuleNotFoundError:
    pass


def predict_tophat_mask(
    score_volume: npt.NDArray[float],
    output_path: pathlib.Path | None = None,
    n_false_positives: float = 1.0,
    create_plot: bool = True,
    tophat_connectivity: int = 1,
    bins: int = 50,
) -> npt.NDArray[bool]:
    """This function gets as input a score map and returns a peak mask as determined
    with a tophat transform.

    It does the following things:
     - calculate a tophat transform using scipy.ndimage.white_tophat() and a kernel
     ndimage.generate_binary_structure(rank=3, connectivity=1).
     - calculate a histogram of the transformed score map and take its log to focus more
        on small values
     - take second derivative of log(histogram) to find the region for fitting a
        Gaussian, where the second derivative switches from negative to positive the
        background noise likely breaks
     - use formula from excellent work of Rickgauer et al. (2017, eLife) which uses the
        error function to find the likelihood of false positives on the background
        Gaussian distribution: N**(-1) = erfc( theta / ( sigma * sqrt(2) ) ) / 2

    Parameters
    ----------
    score_volume: npt.NDArray[float]
        template matching score map
    output_path: Optional[pathlib.Path], default None
        if provided (and plotting is available), write a figure of the fit to the output
        folder
    n_false_positives: float, default 1.0
        number of false positive for error function cutoff calculation
    create_plot: bool, default True
        whether to plot the gaussian fit and cut-off estimation
    tophat_connectivity: int, default 1
        connectivity of binary structure
    bins: int, default 50
        number of bins to use for the historgram for estimation and plotting
    Returns
    -------
    peak_mask: npt.NDArray[bool]
        boolean mask with tophat filtered peak locations
    """
    # convert float16 to float32 otherwise ndimage.white_tophat() fails and
    # score_volume.std() might produce np.inf
    if score_volume.dtype == np.float16:
        score_volume = score_volume.astype(np.float32)

    tophat = ndimage.white_tophat(
        score_volume,
        structure=ndimage.generate_binary_structure(
            rank=3, connectivity=tophat_connectivity
        ),
    )
    y, bins = np.histogram(tophat.flatten(), bins=bins)
    bin_centers = (bins[:-1] + bins[1:]) / 2
    x_raw, y_raw = (
        bin_centers[2:],
        y[2:],
    )  # discard first two points because of over-representation of zeros

    # take second derivative and discard inaccurate boundary value (hence the [2:])
    with np.errstate(divide="ignore"):
        y_log = np.log(y_raw)
        y_log[np.isinf(y_log)] = 0
    second_derivative = np.gradient(np.gradient(y_log))[2:]
    m1 = second_derivative[:-1] < 0  # where the derivative is negative
    sign = np.sign(second_derivative[1:] * second_derivative[:-1])
    # if there was a 0 in second derivative, there are now two 0s in sign
    # replace the 0,0 with the next value and previous value instead
    sign = np.where(sign == 0, np.roll(sign, -1), sign)
    sign = np.where(sign == 0, np.roll(sign, 1), sign)
    m2 = sign == -1  # switches from neg. to pos. and vice versa
    idx = (
        int(np.argmax(m1 & m2))  # first switch from negative to positive
        + 2  # +2 for 2nd derivative discarded part
        + 1  # +1 to include last value
    )
    x_fit, y_fit = x_raw[:idx], y_raw[:idx]

    def gauss(x, amp, mu, sigma):  # gaussian for fitting
        return amp * np.exp(-((x - mu) ** 2) / (2 * sigma**2))

    def log_gauss(x, amp, mu, sigma):  # log of gaussian for fitting
        return np.log(gauss(x, amp, mu, sigma))

    guess = np.array(
        [y.max(), 0, score_volume.std()]
    )  # rough initial guess for amp, mu and sigma
    coeff = curve_fit(gauss, x_fit, y_fit, p0=guess)[
        0
    ]  # first fit regular gauss to get better guesses
    coeff_log = curve_fit(log_gauss, x_fit, np.log(y_fit), p0=coeff)[
        0
    ]  # now go for accurate fit to log of gauss
    search_space = coeff_log[0] / (coeff_log[2] * np.sqrt(2 * np.pi))
    # formula Rickgauer et al. (2017, eLife):
    # N**(-1) = erfc( theta / ( sigma * sqrt(2) ) ) / 2
    # we need to find theta (i.e. the cut-off)
    cut_off = (
        erfcinv((2 * n_false_positives) / search_space) * np.sqrt(2) * coeff_log[2]
        + coeff_log[1]
    )

    if plotting_available and output_path is not None and create_plot:
        fig, ax = plt.subplots()
        ax.scatter(x_raw, y_raw, label="tophat", marker="o")
        ax.plot(x_raw, gauss(x_raw, *coeff_log), label="pred", color="tab:orange")
        ax.axvline(cut_off, color="gray", linestyle="dashed", label="cut-off")
        ax.axvspan(x_fit[0], x_fit[-1], alpha=0.25, color="gray", label="fitted data")
        ax.set_yscale("log")
        ax.set_ylim(bottom=0.1)
        ax.set_ylabel("Occurence")
        ax.set_xlabel("Tophat scores")
        ax.legend()
        plt.tight_layout()
        plt.savefig(output_path, dpi=600, transparent=False, bbox_inches="tight")

    peak_mask = tophat > cut_off

    return peak_mask


def extract_particles(
    job: TMJob,
    n_particles: int,
    particle_diameter: float | None = None,
    cut_off: float | None = None,
    n_false_positives: float = 1.0,
    tomogram_mask_path: pathlib.Path | None = None,
    tophat_filter: bool = False,
    create_plot: bool = True,
    tophat_connectivity: int = 1,
    relion5_compat: bool = False,
    ignore_tomogram_mask: bool = False,
    tophat_bins: int = 50,
    plot_bins: int = 20,
) -> tuple[pd.DataFrame, list[float, ...]]:
    """
    Parameters
    ----------
    job: pytom_tm.tmjob.TMJob
        template matching job to annotate particles from
    n_particles: int
        maximum number of particles to extract
    particle_diameter: Optional[float]
        particle diameter to remove peaks with after a score has been annotated.
        minimal peak-to-peak distance after extraction will be diameter/2.
    cut_off: Optional[float]
        manually override the automated score cut-off estimation, value between 0 and 1
    n_false_positives: float, default 1.0
        tune the number of false positives to be included for automated error function
        cut-off estimation: should be a float > 0
    tomogram_mask_path: Optional[pathlib.Path]
        path to a tomographic binary mask for extraction, will override
        job.tomogram_mask
    tophat_filter: bool
        attempt to only select sharp peaks with the tophat filter
    create_plot: bool, default True
        flag for creating extraction plots
    tophat_connectivity: int, default 1
        connectivity of kernel for tophat transform
    relion5_compat: bool, default False
        relion5 compatibility writes coordinates relative to center and in Angstrom
        center definition should be: tomo_shape / 2 - 1
    ignore_tomogram_mask: bool, default False
        Debug option to force the code to ignore job.tomogram_mask and input mask.
        Allows for re-exctraction without rerunning the TM job
        (assuming the scores volume seems reasonable)
    tophat_bins: int, default 50
        The numbers of bins to use in the tophat histogram
    plot_bins: int, default 20
        The numbers of bins to use for the plot histogram of occurences

    Returns
    -------
    dataframe, scores: tuple[pd.DataFrame, list[float, ...]]
        dataframe with annotations that can be written out as a STAR file and a list of
        the selected scores
    """

    score_volume = read_mrc(job.output_dir.joinpath(f"{job.tomo_id}_scores.mrc"))
    angle_volume = read_mrc(job.output_dir.joinpath(f"{job.tomo_id}_angles.mrc"))
    angle_list = get_angle_list(
        job.rotation_file,
        sort_angles=version.parse(job.pytom_tm_version_number) > version.parse("0.3.0"),
        symmetry=job.rotational_symmetry,
    )

    if tophat_filter:  # constrain the extraction with a tophat filter
        predicted_peaks = predict_tophat_mask(
            score_volume,
            output_path=job.output_dir.joinpath(f"{job.tomo_id}_tophat_filter.svg"),
            n_false_positives=n_false_positives,
            create_plot=create_plot,
            tophat_connectivity=tophat_connectivity,
            bins=tophat_bins,
        )
        score_volume *= (
            predicted_peaks  # multiply with predicted peaks to keep only those
        )

    # apply tomogram mask if provided
    tomogram_mask = None
    if ignore_tomogram_mask:
        logging.warning("Ignoring tomogram mask")
    elif tomogram_mask_path is not None:
        tomogram_mask = read_mrc(tomogram_mask_path)
    elif job.tomogram_mask is not None:
        tomogram_mask = read_mrc(job.tomogram_mask)

    if tomogram_mask is not None:
        if tomogram_mask.shape != job.tomo_shape:
            raise ValueError(
                "Tomogram mask does not have the same number of pixels as the "
                f"tomogram.\n Tomogram mask shape: {tomogram_mask.shape}, "
                f"tomogram shape: {job.tomo_shape}"
            )
        slices = [
            slice(origin, origin + size)
            for origin, size in zip(job.search_origin, job.search_size)
        ]
        tomogram_mask = tomogram_mask[*slices]

        # If tomogram_mask is stored as signed bytes we need to convert it to unsigned:
        if (
            tomogram_mask.dtype == "int8"
            and -128 <= tomogram_mask.min()
            and tomogram_mask.max() <= -127
        ):
            tomogram_mask = (tomogram_mask > -128).astype("int8")

        # mask should be larger than zero in regions of interest!
        score_volume[tomogram_mask <= 0] = 0

    if particle_diameter is not None:
        particle_radius_px = int((particle_diameter / 2) / job.voxel_size)
    elif job.particle_diameter is not None:
        particle_radius_px = int((job.particle_diameter / 2) / job.voxel_size)
        logging.info(
            "No particle diameter was provided, so using the diameter "
            "specified previously to mask out areas around peaks. Take care for "
            "strongly elongated particles as it might prevent correct "
            "annotation when they arrange parallel to each other and close together."
        )
    else:
        raise ValueError(
            "You need to specify a particle diameter to mask out areas around each "
            "peak during extraction!"
        )

    # mask edges of score volume
    score_volume[0:particle_radius_px, :, :] = 0
    score_volume[:, 0:particle_radius_px, :] = 0
    score_volume[:, :, 0:particle_radius_px] = 0
    score_volume[-particle_radius_px:, :, :] = 0
    score_volume[:, -particle_radius_px:, :] = 0
    score_volume[:, :, -particle_radius_px:] = 0

    sigma = job.job_stats["std"]
    search_space = job.job_stats["search_space"]
    if cut_off is None:
        # formula Rickgauer et al. (2017, eLife):
        # N**(-1) = erfc( theta / ( sigma * sqrt(2) ) ) / 2
        # we need to find theta (i.e. the cut off)
        cut_off = erfcinv((2 * n_false_positives) / search_space) * np.sqrt(2) * sigma
        logging.info(f"cut off for particle extraction: {cut_off}")
    elif cut_off < 0:
        logging.warning(
            "Provided extraction score cut-off is smaller than 0. Changing to 0 as "
            "that is smallest allowed value."
        )
        cut_off = 0

    # mask for iteratively selecting peaks
    cut_box = int(particle_radius_px) * 2 + 1
    cut_mask = (spherical_mask(cut_box, particle_radius_px, cut_box // 2) == 0) * 1

    # data for star file
    pixel_size = job.voxel_size
    tomogram_id = job.tomo_id

    # remove relion5 reconstructed tomogram name as it messes with linking the tilt
    # series id when extracting subtomos
    if relion5_compat and tomogram_id.startswith("rec_"):
        tomogram_id = tomogram_id[4:]

    data = []
    scores = []

    for _ in tqdm(range(n_particles)):
        ind = np.unravel_index(score_volume.argmax(), score_volume.shape)

        lcc_max = score_volume[ind]

        if lcc_max <= cut_off:
            break

        scores.append(lcc_max)

        # According to https://www.ccpem.ac.uk/user_help/rotation_conventions.php
        # relion uses clockwise ZYZ. The -1 multiplication is needed because we use
        # anti-clockwise angles.,
        rotation = convert_euler(
            [-1 * a for a in angle_list[int(angle_volume[ind])]],
            order_in="ZXZ",
            order_out="ZYZ",
            degrees_in=False,
            degrees_out=True,
        )

        location = [i + o for i, o in zip(job.search_origin, ind)]

        data.append(
            (
                location[0],  # CoordinateX
                location[1],  # CoordinateY
                location[2],  # CoordinateZ
                rotation[0],  # AngleRot
                rotation[1],  # AngleTilt
                rotation[2],  # AnglePsi
                lcc_max,  # LCCmax
                cut_off,  # Extraction cut off
                sigma,  # Add sigma of template matching search, LCCmax/sigma = SNR
                pixel_size,  # DetectorPixelSize
                tomogram_id,  # MicrographName
            )
        )

        # box out the particle
        start = [i - particle_radius_px for i in ind]
        score_volume[
            start[0] : start[0] + cut_box,
            start[1] : start[1] + cut_box,
            start[2] : start[2] + cut_box,
        ] *= cut_mask

    output = pd.DataFrame(
        data,
        columns=[
            "rlnCoordinateX",
            "rlnCoordinateY",
            "rlnCoordinateZ",
            "rlnAngleRot",
            "rlnAngleTilt",
            "rlnAnglePsi",
            "rlnLCCmax",
            "rlnCutOff",
            "rlnSearchStd",
            "rlnDetectorPixelSize",
            "rlnMicrographName",
        ],
    )

    if relion5_compat:
        dims = np.array(job.tomo_shape)
        if isinstance(job.ts_metadata, RelionTiltSeriesMetaData):
            binning = job.ts_metadata.binning
            center = (dims * binning) / 2 - 1
            voxel_size = job.ts_metadata.tilt_series_pixel_size
        else:
            # we approximate the center, the correct way is to use
            # relion5's unbinned dimensions and calculate:
            # center = (ubinned_dims / 2) - 1 as above
            binning = 1
            center = dims / 2
            voxel_size = job.voxel_size
        output["rlnCoordinateX"], output["rlnCoordinateY"], output["rlnCoordinateZ"] = (
            (output["rlnCoordinateX"] * binning - center[0]) * voxel_size,
            (output["rlnCoordinateY"] * binning - center[1]) * voxel_size,
            (output["rlnCoordinateZ"] * binning - center[2]) * voxel_size,
        )
        column_change = {
            "rlnCoordinateX": "rlnCenteredCoordinateXAngst",
            "rlnCoordinateY": "rlnCenteredCoordinateYAngst",
            "rlnCoordinateZ": "rlnCenteredCoordinateZAngst",
            "rlnMicrographName": "rlnTomoName",
            "rlnDetectorPixelSize": "rlnTomoTiltSeriesPixelSize",
        }
        output = output.rename(columns=column_change)

    if plotting_available and create_plot:
        y, bins = np.histogram(scores, bins=plot_bins)
        x = (bins[1:] + bins[:-1]) / 2
        hist_step = bins[1] - bins[0]
        # add more starting values for background Gaussian
        x_ext = np.concatenate((np.linspace(x[0] - 5 * hist_step, x[0], 10), x))
        # calculate amplitude of Gaussian and correct for histogram step size
        noise_amplitude = (search_space / (sigma * np.sqrt(2 * np.pi))) * hist_step
        y_background = noise_amplitude * np.exp(-(x_ext**2) / (2 * sigma**2))

        fig, ax = plt.subplots()
        ax.scatter(x, y, label="extracted", marker="o")
        ax.plot(x_ext, y_background, label="background", color="tab:orange")
        ax.axvline(cut_off, color="gray", linestyle="dashed", label="cut-off")
        ax.set_ylim(bottom=0, top=2 * max(y))
        ax.set_ylabel("Occurence")
        ax.set_xlabel(r"${LCC}_{max}$")
        ax.legend()
        plt.tight_layout()
        plt.savefig(
            job.output_dir.joinpath(f"{job.tomo_id}_extraction_graph.svg"),
            dpi=600,
            transparent=False,
            bbox_inches="tight",
        )

    return output, scores



import pandas as pd
import starfile
import pathlib
import logging


def merge_stars(
    input_star_files: list[pathlib.Path],
    output_file: pathlib.Path,
    relion5_compat: bool = False,
) -> None:
    """Merge particle starfiles together for RELION4 or make an importable RELION5
    starfile pointing to all the starfiles

    Parameters
    ----------
    input_star_file: list[pathlib.Path]
        List of input starfiles
    output_file: pathlib.Path
        Output file to be written
    relion5_compat: Bool, default False
        If True, write the new RELION5 import type starfile instead of just
        concatenating all the starfiles together (default)
    """
    # Make sure all paths are absolute and unique
    files = set(f.resolve() for f in input_star_files)

    # Warn if we end up with less files (due to symlinks pointing to the same thing
    # or the user giving the same star file multiple times)
    if len(files) != len(input_star_files):
        logging.warning("Found duplicate input, only using unique star files")

    if len(files) == 0:
        raise ValueError("No starfiles in directory.")
    elif len(files) < 2 and not relion5_compat:
        raise ValueError(
            "Only one (unique) starfile given which doesn't make sense to merge"
        )

    def capture_read(f, relion5_compat=False):
        out = starfile.read(f)
        if type(out) is pd.DataFrame:
            return out
        # Assuming dict here
        if not relion5_compat:
            logging.warning(
                f"{f} seems to be a multi-data-block starfile, will only "
                "concatenate the 'particles' data block "
            )
        if "particles" not in out:
            raise ValueError(f"{f} does not have a 'particles' data block")
        return out["particles"]

    if not relion5_compat:
        dataframes = (capture_read(f) for f in files)
        logging.info("Concatting and writing star files")
        output = pd.concat(dataframes, ignore_index=True)
    else:
        logging.info("Writing out 2-column relion5 star file")
        data = []
        for fname in files:
            df = capture_read(fname, relion5_compat=True)
            if "rlnTomoName" not in df.columns:
                raise ValueError(
                    f"Could not find 'rlnTomoName' column in the file: {fname}. "
                    "Are you sure this is a relion5 star file?"
                )
            for name in set(df["rlnTomoName"]):
                data.append((name, fname))
        output = pd.DataFrame(
            data, columns=["rlnTomoName", "rlnTomoImportParticleFile"]
        )

    starfile.write({"particles": output}, output_file, overwrite=True)



import unittest
import numpy as np
from pytom_tm.extract import predict_tophat_mask


class TestExtract(unittest.TestCase):
    def test_predict_tophat_mask(self):
        rng = np.random.default_rng(0)
        volume = rng.normal(loc=0, scale=0.1, size=(50,) * 3)  # generate random peaks
        volume[20, 20, 20] = 1
        tophat_mask = predict_tophat_mask(volume)
        self.assertEqual(
            tophat_mask.shape,
            volume.shape,
            msg="tophat mask should have same size as input",
        )
        self.assertEqual(
            tophat_mask.dtype, bool, msg="predicted tophat mask should be boolean"
        )
        self.assertNotEqual(tophat_mask.sum(), 0, msg="tophat mask should not be empty")

        # test with float16 scores (internally converted to float32)
        tophat_mask = predict_tophat_mask(volume.astype(np.float16))
        self.assertNotEqual(
            tophat_mask.sum(), 0, msg="float16 scores failing for tophat mask"
        )

    # part of the extraction test in test_tmjob.py
    # def test_extract_job_with_tomogram_mask(self):
    #    pass



from dataclasses import dataclass, replace
from pytom_tm.json import JsonSerializable


@dataclass(kw_only=True)
class CtfData(JsonSerializable):
    """A data class to combine all CTF data for a single tilt

    Attributes
    ----------
    defocus: float
        defocus in m
    amplitude_contrast: float
        fraction of amplitude contrast between 0 and 1
    voltage: float
        voltage in eV
    spherical_aberration: float
        spherical aberration in m
    phase_shift_deg: float, default 0.0
        phase shift for phase plates in degree
    flip_phase: bool, default False
        wether we should apply the phase-flip ctf correction
    """

    defocus: float
    amplitude_contrast: float
    voltage: float
    spherical_aberration: float
    phase_shift_deg: float = 0.0
    flip_phase: bool = False

    def replace(self, **kwargs):
        return replace(self, **kwargs)


@dataclass(kw_only=True, frozen=True)
class TiltSeriesMetaData(JsonSerializable):
    """A dataclass to keep all tilt-series metadata together

    Attributes
    ----------
    tilt_angles: list[float]
        list of tilt angles of the tilt-series in degrees
    angles_in_degrees: bool, default True
        Wether the tilt angles are in degree (default) or radian
    ctf_data: list[CtfData] | None, default None
        list of CtfData per tilt, should either be length 1 if it is identical for all
        tilts or the same length as tilt_angles
    dose_accumulation: list[float] | None, default None
        list of accumulated doses per tilt in electrons per Å^2,
        if given should have the same length as tilt_angles
    defocus_handedness: int from {-1, 0, 1}, default 0
        defocus handesness for gradient correction
        as specified in Pyle and Zianetti (2021)
        0: no defocus gradient correction (default),
        1: correction assuming correct handedness
       -1: the handedness will be inverted
    per_tilt_weighting: bool, default False
        if we want to do per-tilt weighting to create a fanned wedge instead
        of a default binary one
    """

    tilt_angles: list[float]
    angles_in_degrees: bool = True
    ctf_data: list[CtfData] | None = None
    dose_accumulation: list[float] | None = None
    defocus_handedness: int = 0
    per_tilt_weighting: bool = False

    def replace(self, **kwargs):
        return replace(self, **kwargs)

    def __post_init__(self):
        if not isinstance(self.tilt_angles, list) or len(self.tilt_angles) < 2:
            raise ValueError("TiltSeriesMetaData requires at least 2 tilt angles")
        n_angles = len(self.tilt_angles)
        # Make sure all lists have the same value
        if self.ctf_data is not None and len(self.ctf_data) == 1:
            # Weird invocation needed to circumvent frozen disallowing direct assignment
            object.__setattr__(self, "ctf_data", self.ctf_data * n_angles)
        elif self.ctf_data is not None and len(self.ctf_data) != n_angles:
            raise ValueError(
                "Expected either a list with a single CtfData or the "
                f"same number as tilt_angles ({n_angles}). Got {len(self.ctf_data)} "
                "instead."
            )
        if (
            self.dose_accumulation is not None
            and len(self.dose_accumulation) != n_angles
        ):
            raise ValueError(
                "Expected a list with the same number of doses as tilt "
                f"angles ({n_angles}). Got {len(self.dose_accumulation)} instead."
            )

        if self.defocus_handedness not in {-1, 0, 1}:
            raise ValueError(
                "Got an invalid defocus handedness, "
                f"expected -1, 0, or 1. Got: {self.defocus_handedness}"
            )

    def __len__(self):
        return len(self.tilt_angles)


@dataclass(kw_only=True, frozen=True)
class WarpTiltSeriesMetaData(TiltSeriesMetaData):
    """Extension of TiltSeriesMetaData that has per_tilt_weighting default as True"""

    per_tilt_weighting: bool = True


@dataclass(kw_only=True, frozen=True)
class RelionTiltSeriesMetaData(TiltSeriesMetaData):
    """Extension of the TiltSeriesMetaData that has per_tilt_weighting default as True
    and has the following extra attributes

    Attributes
    ----------
    binning: float
        binning of the tomogram
    tilt_series_pixel_size: float
        pixel size of the original tilt series in Å
    """

    binning: float
    tilt_series_pixel_size: float
    per_tilt_weighting: bool = True


# TODO: possible TomogramMetaData dataclass
# @dataclass(kw_only=True)
# class TomogramMetaData(JsonSerializable):
#    """A dataclass for to keep all the meta-data for the tomogram together
#
#    """
#    path: pathlib.Path
#    voxel_size: float | None = None
#    ts_metadata: TiltSeriesMetaData = None
#    mask: pathlib.Path | None = None
#    #TODO: add mrc header info??
#
#    def __post_init__(self):
#        # Set and check the voxel size if not given



import unittest
import sys
import pathlib
import numpy as np
import cupy as cp
import logging
from shutil import which
from contextlib import redirect_stdout, redirect_stderr
from io import StringIO
from tempfile import TemporaryDirectory
from pytom_tm import entry_points
from pytom_tm import io
from testing_utils import chdir

# (command line function, function in entry_points file)
ENTRY_POINTS_TO_TEST = [
    ("pytom_create_mask.py", "pytom_create_mask"),
    ("pytom_create_template.py", "pytom_create_template"),
    ("pytom_match_template.py", "match_template"),
    ("pytom_extract_candidates.py", "extract_candidates"),
    ("pytom_merge_stars.py", "merge_stars"),
]
# Test if optional dependencies are installed
try:
    from pytom_tm import plotting  # noqa: F401
except RuntimeError:
    pass
else:
    ENTRY_POINTS_TO_TEST.append(("pytom_estimate_roc.py", "estimate_roc"))

# Input files for command line match_template
TEST_DATA = pathlib.Path(__file__).parent.joinpath("test_data")
TEMPLATE = TEST_DATA.joinpath("template.mrc")
MASK = TEST_DATA.joinpath("mask.mrc")
TOMOGRAM = TEST_DATA.joinpath("tomogram.mrc")
TILT_ANGLES = TEST_DATA.joinpath("angles.rawtlt")
TILT_ANGLES_MULTI_COLUMN = TEST_DATA.joinpath("angles_multi_column.rawtlt")
DOSE = TEST_DATA.joinpath("test_dose.txt")
DEFOCUS = TEST_DATA.joinpath("defocus.txt")
DEFOCUS_IMOD = (
    pathlib.Path(__file__).parent.joinpath("Data").joinpath("test_imod.defocus")
)
WARP_XML = pathlib.Path(__file__).parent.joinpath(
    "Data/warptools_xml_example/gs04_ts_003.xml"
)

RELION5_TOMOGRAMS_STAR = pathlib.Path(__file__).parent.joinpath(
    "Data/relion5_project_example/Tomograms/job009/tomograms.star"
)
RELION5_TOMOGRAM = TEST_DATA.joinpath("rec_tomo200528_107.mrc")

# Initial logging level
LOG_LEVEL = logging.getLogger().level


def prep_argv(arg_dict):
    argv = []
    [
        argv.extend([k] + v.split()) if v != "" else argv.append(k)
        for k, v in arg_dict.items()
    ]
    return argv


class TestParseArgv(unittest.TestCase):
    def test_parse_argv(self):
        out = entry_points._parse_argv()
        # test behavior by repeating the behavior
        self.assertEqual(sys.argv[1:], out)
        inp = ["test1", "test2"]
        out = entry_points._parse_argv(inp)
        for i in inp:
            self.assertIn(i, out)


class TestEntryPoints(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        TEST_DATA.mkdir(parents=True)
        io.write_mrc(TEMPLATE, np.zeros((5, 5, 5), dtype=np.float32), 1)
        io.write_mrc(MASK, np.zeros((5, 5, 5), dtype=np.float32), 1)
        io.write_mrc(TOMOGRAM, np.zeros((10, 10, 10), dtype=np.float32), 1)
        io.write_mrc(RELION5_TOMOGRAM, np.zeros((10, 10, 10), dtype=np.float32), 1)
        np.savetxt(TILT_ANGLES, np.linspace(-50, 50, 35))
        np.savetxt(
            TILT_ANGLES_MULTI_COLUMN,
            np.array((np.linspace(-50, 50, 35), np.arange(35))).T,
            fmt=["%.18e", "%d"],
        )
        np.savetxt(DOSE, np.linspace(0, 100, 35))
        np.savetxt(DEFOCUS, np.ones(35) * 3000)

    @classmethod
    def tearDownClass(cls) -> None:
        TEMPLATE.unlink()
        MASK.unlink()
        TOMOGRAM.unlink()
        RELION5_TOMOGRAM.unlink()
        TILT_ANGLES.unlink()
        TILT_ANGLES_MULTI_COLUMN.unlink()
        DOSE.unlink()
        DEFOCUS.unlink()
        TEST_DATA.rmdir()

    def setUp(self):
        # set up a destination temp dir
        tempdir = TemporaryDirectory()
        self.outputdir = pathlib.Path(tempdir.name)
        self.addCleanup(tempdir.cleanup)

    def test_entry_points_exist(self):
        for cli, fname in ENTRY_POINTS_TO_TEST:
            # test the command line function can be found
            self.assertIsNotNone(which(cli))
            # assert the entry_point be called with -h and exit cleanly
            # catch stdout to prevent shell polution
            func = getattr(entry_points, fname)
            dump = StringIO()
            with self.assertRaises(SystemExit) as ex, redirect_stdout(dump):
                func([cli, "-h"])
            dump.close()
            # check if the system return code is 0 (success)
            self.assertEqual(ex.exception.code, 0)

    def test_create_mask(self):
        defaults = {
            "-b": "60",
            "-r": "12",
        }
        default_output_name = f"mask_b{defaults['-b']}px_r{float(defaults['-r'])}px.mrc"

        def start(arg_dict):
            entry_points.pytom_create_mask(prep_argv(arg_dict))

        # Test defaults, do change to temp dir
        with chdir(self.outputdir):
            start(defaults)
            # Make sure default file exists
            self.assertTrue(pathlib.Path(default_output_name).exists())
            # Make sure it has the expected size
            meta_data = io.read_mrc_meta_data(default_output_name)
            self.assertEqual(len(meta_data["shape"]), 3)
            for n in meta_data["shape"]:
                self.assertEqual(n, 60)

        # Check we fail loud if only one of the two options is given
        for i in ["--radius-minor1", "--radius-minor2"]:
            inp = defaults.copy()
            inp[i] = "6"
            with self.assertRaisesRegex(ValueError, f"Only got {i}"):
                start(inp)

        # smoke test eliptical mask and uneven box
        inp = defaults.copy()
        inp["--radius-minor1"] = "6"
        inp["--radius-minor2"] = "8"
        inp["-b"] = "55"
        inp["-o"] = str(self.outputdir / "mask_ellipse.mrc")
        start(inp)

        self.assertTrue(self.outputdir.joinpath("mask_ellipse.mrc").exists())
        meta_data = io.read_mrc_meta_data(self.outputdir.joinpath("mask_ellipse.mrc"))
        self.assertEqual(len(meta_data["shape"]), 3)
        for n in meta_data["shape"]:
            self.assertEqual(n, 55)

    def test_create_template(self):
        defaults = {"-i": str(TEMPLATE), "--output-voxel-size-angstrom": "1.000"}
        default_output_name = f"template_{TEMPLATE.stem}_{float(1)}A.mrc"

        def start(arg_dict):
            entry_points.pytom_create_template(prep_argv(arg_dict))

        # Test defaults, do change to temp dir
        with chdir(self.outputdir):
            start(defaults)
            # Make sure default file exists
            self.assertTrue(pathlib.Path(default_output_name).exists())
            # Make sure it has the expected size
            meta_data = io.read_mrc_meta_data(default_output_name)
            self.assertEqual(len(meta_data["shape"]), 3)
            for n in meta_data["shape"]:
                self.assertEqual(n, 5)
            self.assertEqual(meta_data["voxel_size"], 1.0)

        # Test warning and no warning on input voxelsize
        args = defaults.copy()
        output = self.outputdir / "rounded_template.mrc"
        args["-o"] = f"{output}"
        # should round correctly at 3 digits
        args["--input-voxel-size"] = "0.9999"
        args["--log-test"] = ""
        with self.assertNoLogs(level=logging.WARNING):
            start(args)
        self.assertTrue(output.exists())

        # Now test with the warning
        args = defaults.copy()
        output = self.outputdir / "wrong_rounded_template.mrc"
        args["-o"] = f"{output}"
        # use 6 as 5 still rounds corrctly due to python bankers' round
        args["--input-voxel-size"] = "1.0006"
        # Don't try to invent pixels
        args["--output-voxel-size-angstrom"] = "2.0"
        args["--log-test"] = ""
        with self.assertLogs(level="WARNING") as cm:
            start(args)
        self.assertEqual(len(cm.output), 1)
        self.assertIn("voxel size does not match", cm.output[0])
        self.assertTrue(output.exists())

        # Test failure on trying to get more resolution
        args = defaults.copy()
        args["--output-voxel-size-angstrom"] = "0.5"
        with self.assertRaisesRegex(NotImplementedError, "smaller voxel size"):
            start(args)

    def test_match_template(self):
        defaults = {
            "-t": str(TEMPLATE),
            "-m": str(MASK),
            "-v": str(TOMOGRAM),
            "-d": str(self.outputdir),
            "--angular-search": "35",
            "--tilt-angles": str(TILT_ANGLES),
            "--per-tilt-weighting": "",
            "--dose-accumulation": str(DOSE),
            "--defocus": str(DEFOCUS_IMOD),
            "--amplitude-contrast": "0.08",
            "--spherical-aberration": "2.7",
            "--voltage": "300",
            "--tomogram-ctf-model": "phase-flip",
            "-g": "0",
        }

        def start(arg_dict):  # simplify run
            entry_points.match_template(prep_argv(arg_dict))

        # test valid defocus arguments
        for z in [str(DEFOCUS_IMOD), str(DEFOCUS), "3000"]:
            arguments = defaults.copy()
            arguments["--defocus"] = z
            start(arguments)

        # test faulty args
        for z in ["asdf.txt", "asdf"]:
            dump = StringIO()
            with (
                self.assertRaises(SystemExit) as ex,
                redirect_stdout(dump),
                redirect_stderr(dump),
            ):
                arguments = defaults.copy()
                arguments["--defocus"] = z
                start(arguments)
            dump.close()
            # check if the system return code is 0 (success)
            self.assertEqual(ex.exception.code, 2)

        # remove per-tilt-weighting
        arguments = defaults.copy()
        arguments.pop("--per-tilt-weighting")
        start(arguments)

        with self.assertRaises(
            ValueError, msg="Missing CTF params should produce error"
        ):
            arguments = defaults.copy()
            arguments.pop("--voltage")
            start(arguments)

        # test angular search and particle diameter options
        with self.assertRaises(
            ValueError, msg="Missing angular search should raise an error."
        ):
            arguments = defaults.copy()
            arguments.pop("--angular-search")
            start(arguments)

        arguments = defaults.copy()
        arguments.pop("--angular-search")
        arguments["--particle-diameter"] = "50"
        # set low-pass to tune the search to lower degree
        arguments["--low-pass"] = "50"
        start(arguments)

        # phase randomization test
        arguments = defaults.copy()
        arguments["-r"] = ""
        start(arguments)
        # test if we can set the rng seed, see issue #194
        arguments["--rng-seed"] = "42"
        start(arguments)

        # test debug files
        arguments = defaults.copy()
        arguments["--log"] = "debug"
        start(arguments)
        # these files will only exist if the test managed to set the logging correctly
        self.assertTrue(
            self.outputdir.joinpath("template_psf.mrc").exists(),
            msg="File should exist in debug mode",
        )
        self.assertTrue(
            self.outputdir.joinpath("template_convolved.mrc").exists(),
            msg="File should exist in debug mode",
        )

        # reset the log level after the entry point modified it
        logging.basicConfig(level=LOG_LEVEL, force=True)

        # test providing invalid gpu indices
        n_devices = cp.cuda.runtime.getDeviceCount()
        for indices in ["-1", f"0 {n_devices}"]:
            dump = StringIO()
            with (
                self.assertRaises(SystemExit) as ex,
                redirect_stdout(dump),
                redirect_stderr(dump),
            ):
                arguments = defaults.copy()
                arguments["-g"] = indices
                start(arguments)
            self.assertIn("gpu indices", dump.getvalue())
            dump.close()
        # test error when volume splits can't be evenly distributed by gpus
        arguments = defaults.copy()
        # 4 pieces can't be fit on 3 gpus
        arguments["--volume-split"] = "2 2 1"
        arguments["-g"] = "0 0 0"
        with self.assertRaisesRegex(ValueError, r"4 tomogram pieces.*3 GPUs"):
            start(arguments)

        # test warp xml metadata reading
        arguments = defaults.copy()
        [
            arguments.pop(x)
            for x in [
                "--tilt-angles",
                "--per-tilt-weighting",
                "--dose-accumulation",
                "--defocus",
                "--amplitude-contrast",
                "--spherical-aberration",
                "--voltage",
            ]
        ]
        arguments["-v"] = str(RELION5_TOMOGRAM)
        arguments["--warp-xml-file"] = str(WARP_XML)
        start(arguments)

        # test relion5 metadata reading
        arguments = defaults.copy()
        [
            arguments.pop(x)
            for x in [
                "--tilt-angles",
                "--per-tilt-weighting",
                "--dose-accumulation",
                "--defocus",
                "--amplitude-contrast",
                "--spherical-aberration",
                "--voltage",
            ]
        ]
        arguments["-v"] = str(RELION5_TOMOGRAM)
        arguments["--relion5-tomograms-star"] = str(RELION5_TOMOGRAMS_STAR)
        start(arguments)

        with self.assertRaises(
            ValueError, msg="Missing tilt angles should raise an error."
        ):
            arguments.pop("--relion5-tomograms-star")
            start(arguments)

        # test multi_column_angle_files
        # make sure we error out if both options are given
        with self.assertRaises(
            ValueError,
            msg=(
                "Only one of '--tilt-angles' or '--tilt-angles-first-column' is allowed"
            ),
        ):
            arguments = defaults.copy()
            arguments["--tilt-angles-first-column"] = str(TILT_ANGLES_MULTI_COLUMN)
            start(arguments)
        # make sure we error out if multi column angle file is give to regular flag
        dump = StringIO()
        with (
            self.assertRaises(SystemExit) as ex,
            redirect_stdout(dump),
            redirect_stderr(dump),
        ):
            arguments = defaults.copy()
            arguments["--tilt-angles"] = str(TILT_ANGLES_MULTI_COLUMN)
            start(arguments)
        self.assertIn("--tilt-angles-first-column", dump.getvalue())
        dump.close()
        # make sure we can start with the correct option
        arguments = defaults.copy()
        arguments.pop("--tilt-angles")
        arguments["--tilt-angles-first-column"] = str(TILT_ANGLES_MULTI_COLUMN)
        start(arguments)



from __future__ import annotations
from packaging import version
import pathlib
import warnings
import copy
import itertools as itt
import numpy as np
import numpy.typing as npt
import json
import logging
from scipy.fft import next_fast_len, rfftn, irfftn
from pytom_tm.angles import get_angle_list
from pytom_tm.matching import TemplateMatchingGPU
from pytom_tm.weights import (
    create_wedge,
    power_spectrum_profile,
    profile_to_weighting,
    create_gaussian_band_pass,
)
from pytom_tm.io import read_mrc_meta_data, read_mrc, write_mrc, UnequalSpacingError
from pytom_tm.json import CustomJSONEncoder, CustomJSONDecoder
from pytom_tm.dataclass import CtfData, TiltSeriesMetaData, RelionTiltSeriesMetaData
from pytom_tm import __version__ as PYTOM_TM_VERSION


def load_json_to_tmjob(
    file_name: pathlib.Path, load_for_extraction: bool = True
) -> TMJob:
    """Load a previous job that was stored with TMJob.write_to_json().

    Parameters
    ----------
    file_name: pathlib.Path
        path to TMJob json file
    load_for_extraction: bool, default True
        whether a finished job is loaded form disk for extraction, default is True as
        this function is currently only called for pytom_extract_candidates and
        pytom_estimate_roc which run on previously finished jobs

    Returns
    -------
    job: TMJob
        initialized TMJob
    """
    with open(file_name) as fstream:
        data = json.load(fstream, cls=CustomJSONDecoder)

    # Deal with loading ctfdata that was a stored before 0.11.0
    ctf_data = data.get("ctf_data", None)
    if ctf_data is not None and type(ctf_data[0]) is dict:
        data["ctf_data"] = [CtfData(**ctf) for ctf in ctf_data]

    # Deal with loading ts_metadata that was stored before 0.11.0
    if "ts_metadata" not in data:
        kw_dict = {
            "tilt_angles": data["tilt_angles"],
            "ctf_data": data["ctf_data"],
            "dose_accumulation": data.get("dose_accumulation", None),
            "defocus_handedness": data.get("defocus_handedness", 0),
            "per_tilt_weighting": data.get("tilt_weighting"),
        }
        if "metadata" in data:
            # relion5 metadata instead
            kw_dict["binning"] = data["metadata"]["relion5_binning"]
            kw_dict["tilt_series_pixel_size"] = data["metadata"]["relion5_ts_ps"]
            cls = RelionTiltSeriesMetaData
        else:
            cls = TiltSeriesMetaData
        data["ts_metadata"] = cls(**kw_dict)

    # wrangle dtypes
    output_dtype = data.get("output_dtype", "float32")
    output_dtype = np.dtype(output_dtype)

    job = TMJob(
        data["job_key"],
        data["log_level"],
        pathlib.Path(data["tomogram"]),
        pathlib.Path(data["template"]),
        pathlib.Path(data["mask"]),
        pathlib.Path(data["output_dir"]),
        ts_metadata=data["ts_metadata"],
        angle_increment=data.get("angle_increment", data["rotation_file"]),
        mask_is_spherical=data["mask_is_spherical"],
        voxel_size=data["voxel_size"],
        search_x=data["search_x"],
        search_y=data["search_y"],
        search_z=data["search_z"],
        # Use 'get' for backwards compatibility
        tomogram_mask=data.get("tomogram_mask", None),
        low_pass=data["low_pass"],
        # Use 'get' for backwards compatibility
        high_pass=data.get("high_pass", None),
        whiten_spectrum=data.get("whiten_spectrum", False),
        rotational_symmetry=data.get("rotational_symmetry", 1),
        # if version number is not in the .json, it must be 0.3.0 or older
        pytom_tm_version_number=data.get("pytom_tm_version_number", "0.3.0"),
        job_loaded_for_extraction=load_for_extraction,
        particle_diameter=data.get("particle_diameter", None),
        random_phase_correction=data.get("random_phase_correction", False),
        rng_seed=data.get("rng_seed", 321),
        output_dtype=output_dtype,
    )
    job.whole_start = data["whole_start"]
    job.sub_start = data["sub_start"]
    job.sub_step = data["sub_step"]
    job.n_rotations = data["n_rotations"]
    job.start_slice = data["start_slice"]
    job.steps_slice = data["steps_slice"]
    job.job_stats = data["job_stats"]
    return job


def get_defocus_offsets(
    patch_center_x: float,
    patch_center_z: float,
    ts_metadata: TiltSeriesMetaData,
) -> npt.NDArray[float]:
    """Calculate the defocus offsets for a subvolume
    based on the tilt geometry.

    I used the definition from Pyle & Zianetti (https://doi.org/10.1042/BCJ20200715)
    for the default setting of the defocus handedness. It assumes the defocus
    increases for positive tilt angles on the right side of the sample (positive X
    coordinate relative to the center).

    The offset is calculated as follows:
        z_offset = z_center * np.cos(tilt_angle) + x_center * np.sin(tilt_angle)

    Parameters
    ----------
    patch_center_x: float
        x center of subvolume relative to tomogram center
    patch_center_z: float
        z center of subvolume relative to tomogram center
    ts_metadata: TiltSeriesMetaData
        TiltSeriesMetaData containing tilt angles

    Returns
    -------
    z_offsets: npt.NDArray[float]
        an array of defocus offsets for each tilt angle
    """
    n_tilts = len(ts_metadata)
    x_centers = np.full(n_tilts, patch_center_x)
    z_centers = np.full(n_tilts, patch_center_z)
    ta_array = np.array(ts_metadata.tilt_angles)
    if ts_metadata.angles_in_degrees:
        ta_array = np.deg2rad(ta_array)
    if ts_metadata.defocus_handedness < 0:
        ta_array *= -1
    z_offsets = z_centers * np.cos(ta_array) + x_centers * np.sin(ta_array)
    return z_offsets


def _determine_1D_fft_splits(
    length: int, splits: int, overhang: int = 0
) -> list[tuple[tuple[int, int], tuple[int, int]]]:
    """Split a 1D length into FFT optimal sizes taking into account overhangs

    Parameters
    ----------
    length: int
        Total 1D length to split
    splits: int
        Number of splits to make
    overhang: int, default 0
        Minimal overhang/overlap to consider between splits

    Returns
    -------
    output: list[tuple[tuple[int, int], tuple[int,int]]]:
        A list splits where every split gets two tuples meaning:
          [start, end) of the tomogram data in this split
          [start, end) of the unique datapoints in this split
        If a datapoint exists in 2 splits, we add it as unique to
        either the split with the most data or the left one if both
        splits have the same size
    """
    # Everything in this code assumes default slices of [x,y) so including x but
    # excluding y
    data_slices = []
    valid_data_slices = []
    sub_len = []
    # if single split return early
    if splits == 1:
        return [((0, length), (0, length))]
    if splits > length:
        warnings.warn(
            "More splits than pixels were asked, will default to 1 split per pixel",
            RuntimeWarning,
        )
        splits = length
    # Ceil to guarantee that we map the whole length with enough buffer
    min_len = int(np.ceil(length / splits)) + overhang
    min_unique_len = min_len - overhang
    no_overhang_left = 0
    while True:
        if no_overhang_left == 0:
            # Treat first split specially, only right overhang
            split_length = next_fast_len(min_len)
            data_slices.append((0, split_length))
            valid_data_slices.append((0, split_length - overhang))
            no_overhang_left = split_length - overhang
            sub_len.append(split_length)
        elif no_overhang_left + min_unique_len >= length:
            # Last slice, only overhang to the left
            split_length = next_fast_len(min_len)
            data_slices.append((length - split_length, length))
            valid_data_slices.append((length - split_length + overhang, length))
            sub_len.append(split_length)
            break
        else:
            # Any other slice
            split_length = next_fast_len(min_len + overhang)
            left_overhang = (split_length - min_unique_len) // 2
            temp_left = no_overhang_left - left_overhang
            temp_right = temp_left + split_length
            data_slices.append((temp_left, temp_right))
            valid_data_slices.append((temp_left + overhang, temp_right - overhang))
            sub_len.append(split_length)
            no_overhang_left = temp_right - overhang
        if split_length <= 0 or no_overhang_left <= 0:
            raise RuntimeError(
                f"Cannot generate legal splits for {length=}, {splits=}, {overhang=}"
            )
    # Now generate the best unique data point,
    # we always pick the bigest data subset or the left one
    unique_data = []
    unique_left = 0
    for i, (len1, len2) in enumerate(itt.pairwise(sub_len)):
        if len1 >= len2:
            right = valid_data_slices[i][1]
        else:
            right = valid_data_slices[i + 1][0]
        unique_data.append((unique_left, right))
        unique_left = right
    # Add final part
    if unique_left != length:
        unique_data.append((unique_left, length))
    # Make sure unique slices are unique and within valid data
    last_right = 0
    for (vd_left, vd_right), (ud_left, ud_right) in zip(valid_data_slices, unique_data):
        if (
            ud_left < vd_left
            or ud_right > vd_right
            or ud_right > length
            or ud_left != last_right
        ):  # pragma: no cover
            raise RuntimeError(
                f"We produced inconsistent slices for {length=}, {splits=}, {overhang=}"
            )
        last_right = ud_right
    return list(zip(data_slices, unique_data))


class TMJobError(Exception):
    """TMJob Exception with provided message."""

    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super().__init__(message)


class TMJob:
    def __init__(
        self,
        job_key: str,
        log_level: int,
        tomogram: pathlib.Path,
        template: pathlib.Path,
        mask: pathlib.Path,
        output_dir: pathlib.Path,
        ts_metadata: TiltSeriesMetaData,
        voxel_size: float | None = None,
        angle_increment: str | float | None = None,
        mask_is_spherical: bool = True,
        search_x: list[int, int] | None = None,
        search_y: list[int, int] | None = None,
        search_z: list[int, int] | None = None,
        tomogram_mask: pathlib.Path | None = None,
        low_pass: float | None = None,
        high_pass: float | None = None,
        whiten_spectrum: bool = False,
        rotational_symmetry: int = 1,
        pytom_tm_version_number: str = PYTOM_TM_VERSION,
        job_loaded_for_extraction: bool = False,
        particle_diameter: float | None = None,
        random_phase_correction: bool = False,
        rng_seed: int = 321,
        output_dtype: np.dtype = np.float32,
    ):
        """
        Parameters
        ----------
        job_key: str
            job identifier
        log_level: int
            log level for logging module
        tomogram: pathlib.Path
            path to tomogram MRC
        template: pathlib.Path
            path to template MRC
        mask: pathlib.Path
            path to mask MRC
        output_dir: pathlib.Path
            path to output directory
        ts_metadata: TiltSeriesMetaData
            tilt series metadata of the tomogram, contains at least the tilt angles
        voxel_size: float | None, default None
            voxel size of tomogram and template (in A) if not provided will be read from
            template/tomogram MRCs
        angle_increment: Union[str, float]; default 7.00
            angular increment of template search
        mask_is_spherical: bool, default True
            whether template mask is spherical, reduces computation complexity
        search_x: Optional[list[int, int]], default None
            restrict tomogram search region along the x-axis
        search_y: Optional[list[int, int]], default None
            restrict tomogram search region along the y-axis
        search_z: Optional[list[int, int]], default None
            restrict tomogram search region along the z-axis
        tomogram_mask: Optional[pathlib.Path], default None
            when volume splitting tomograms, only subjobs where any(mask > 0) will be
            generated
        low_pass: Optional[float], default None
            optional low-pass filter (resolution in A) to apply to tomogram and template
        high_pass: Optional[float], default None
            optional high-pass filter (resolution in A) to apply to tomogram and
            template
        whiten_spectrum: bool, default False
            whether to apply spectrum whitening
        rotational_symmetry: int, default 1
            specify a rotational symmetry around the z-axis, is only valid if the
            symmetry axis of the template is aligned with the z-axis
        pytom_tm_version_number: str, default current version
            a string with the version number of pytom_tm for backward compatibility
        job_loaded_for_extraction: bool, default False
            flag to set for finished template matching jobs that are loaded back for
            extraction, it prevents recomputation of the whitening filter
        particle_diameter: Optional[float], default None
            particle diameter (in Angstrom) to calculate angular search
        random_phase_correction: bool, default False,
            run matching with a phase randomized version of the template to correct
            scores for noise
        rng_seed: int, default 321
            set a seed for the rng for phase randomization
        output_dtype: np.dtype, default np.float32
            output score volume dtype, options are np.float32 and np.float16
        """
        self.mask = mask
        self.mask_is_spherical = mask_is_spherical
        self.output_dir = output_dir

        self.tomogram = tomogram
        self.template = template
        self.tomo_id = self.tomogram.stem

        self.ts_metadata = ts_metadata

        try:
            meta_data_tomo = read_mrc_meta_data(self.tomogram)
        except UnequalSpacingError:  # add information that the problem is the tomogram
            raise UnequalSpacingError(
                "Input tomogram voxel spacing is not equal in each dimension!"
            )

        try:
            meta_data_template = read_mrc_meta_data(self.template)
        except UnequalSpacingError:  # add information that the problem is the template
            raise UnequalSpacingError(
                "Input template voxel spacing is not equal in each dimension!"
            )

        try:
            meta_data_mask = read_mrc_meta_data(self.mask)
        except UnequalSpacingError:  # add information that the problem is the template
            raise UnequalSpacingError(
                "Input mask voxel spacing is not equal in each dimension!"
            )

        self.tomo_shape = meta_data_tomo["shape"]
        self.template_shape = meta_data_template["shape"]
        self.mask_shape = meta_data_mask["shape"]

        if self.template_shape != self.mask_shape:
            raise ValueError(
                "Template and mask have a different shape in pixels. "
                f"Found template shape: {self.template_shape}. "
                f"Found maks shape: {self.mask_shape}"
            )

        if voxel_size is not None:
            if voxel_size <= 0:
                raise ValueError(
                    "Invalid voxel size provided, smaller or equal to zero."
                )
            self.voxel_size = voxel_size
            if (  # allow tiny numerical differences that are not relevant for
                # template matching
                round(self.voxel_size, 3) != round(meta_data_tomo["voxel_size"], 3)
                or round(self.voxel_size, 3)
                != round(meta_data_template["voxel_size"], 3)
            ):
                logging.debug(
                    f"provided {self.voxel_size} tomogram "
                    f"{meta_data_tomo['voxel_size']} "
                    f"template {meta_data_template['voxel_size']}"
                )
                print(
                    "WARNING: Provided voxel size does not match voxel size annotated "
                    "in tomogram/template mrc."
                )
        elif (
            round(meta_data_tomo["voxel_size"], 3)
            == round(meta_data_template["voxel_size"], 3)
            and meta_data_tomo["voxel_size"] > 0
        ):
            self.voxel_size = round(meta_data_tomo["voxel_size"], 3)
        else:
            raise ValueError(
                "Voxel size could not be assigned, either a mismatch between tomogram "
                "and template or annotated as 0."
            )

        search_origin = [
            x[0] if x is not None else 0 for x in (search_x, search_y, search_z)
        ]
        # Check if tomogram origin is valid
        if all([0 <= x < y for x, y in zip(search_origin, self.tomo_shape)]):
            self.search_origin = search_origin
        else:
            raise ValueError("Invalid input provided for search origin of tomogram.")

        # if end not valid raise and error
        search_end = []
        for x, s in zip([search_x, search_y, search_z], self.tomo_shape):
            if x is not None:
                if not x[1] <= s:
                    raise ValueError(
                        "One of search end indices is larger than the tomogram "
                        "dimension."
                    )
                search_end.append(x[1])
            else:
                search_end.append(s)
        self.search_size = [
            end - start for end, start in zip(search_end, self.search_origin)
        ]

        logging.debug(f"origin, size = {self.search_origin}, {self.search_size}")
        self.tomogram_mask = tomogram_mask
        if tomogram_mask is not None:
            temp = read_mrc(tomogram_mask)
            if temp.shape != self.tomo_shape:
                raise ValueError(
                    "Tomogram mask does not have the same number of pixels as the "
                    "tomogram.\n"
                    f"Tomogram mask shape: {temp.shape}, "
                    f"tomogram shape: {self.tomo_shape}"
                )
            if np.all(temp <= 0):
                raise ValueError(
                    "No values larger than 0 found in the tomogram mask: "
                    f"{tomogram_mask}"
                )

        self.whole_start = None
        # For the main job these are always [0,0,0] and self.search_size, for sub_jobs
        # these will differ from self.search_origin and self.search_size. The main job
        # only uses them to calculate the search_volume_roi for statistics. Sub jobs
        # also use these to extract and place back the relevant region in the master
        # job.
        self.sub_start, self.sub_step = [0, 0, 0], self.search_size.copy()

        # Rotation parameters
        self.start_slice = 0
        self.steps_slice = 1
        self.rotational_symmetry = rotational_symmetry
        self.particle_diameter = particle_diameter
        # calculate increment from particle diameter
        if angle_increment is None:
            if particle_diameter is not None:
                max_res = max(
                    2 * self.voxel_size, low_pass if low_pass is not None else 0
                )
                angle_increment = np.rad2deg(max_res / particle_diameter)
            else:
                angle_increment = 7.0
        self.rotation_file = angle_increment
        try:
            angle_list = get_angle_list(
                angle_increment,
                sort_angles=False,
                symmetry=rotational_symmetry,
                # This log_level is different from self.log_level that is
                # assigned later. The TMJob.log_level refers to the user provided
                # logging setting, while the log_level here is to control the output
                # of the job during candidate extraction/template matching.
                log_level=logging.DEBUG if job_loaded_for_extraction else logging.INFO,
            )
        except ValueError:
            raise TMJobError("Invalid angular search provided.")

        self.n_rotations = len(angle_list)

        # set the band-pass resolution shells
        self.low_pass = low_pass
        self.high_pass = high_pass

        self.whiten_spectrum = whiten_spectrum
        self.whitening_filter = self.output_dir.joinpath(
            f"{self.tomo_id}_whitening_filter.npy"
        )
        if self.whiten_spectrum and not job_loaded_for_extraction:
            logging.info("Estimating whitening filter...")
            weights = 1 / np.sqrt(
                power_spectrum_profile(
                    read_mrc(self.tomogram)[
                        self.search_origin[0] : self.search_origin[0]
                        + self.search_size[0],
                        self.search_origin[1] : self.search_origin[1]
                        + self.search_size[1],
                        self.search_origin[2] : self.search_origin[2]
                        + self.search_size[2],
                    ]
                )
            )
            weights /= weights.max()  # scale to 1
            np.save(self.whitening_filter, weights)

        # phase randomization options
        self.random_phase_correction = random_phase_correction
        self.rng_seed = rng_seed

        # Job details
        self.job_key = job_key
        self.leader = None  # the job that spawned this job
        self.sub_jobs = []  # if this job had no sub jobs it should be executed

        # dict to keep track of job statistics
        self.job_stats = None

        self.log_level = log_level

        # version number of the job
        self.pytom_tm_version_number = pytom_tm_version_number

        # output dtype
        self.output_dtype = output_dtype

    def copy(self) -> TMJob:
        """Create a copy of the TMJob

        Returns
        -------
        job: TMJob
            copied TMJob instance
        """
        return copy.deepcopy(self)

    def write_to_json(self, file_name: pathlib.Path) -> None:
        """Write job to .json file.

        Note: This has to be run from the same cwd as where `self` was initiated
              otherwise the path resolving doesn't make sense

        Parameters
        ----------
        file_name: pathlib.Path
            path to the output file
        """
        d = self.__dict__.copy()
        d.pop("sub_jobs")
        d.pop("search_origin")
        d.pop("search_size")
        d["search_x"] = [
            self.search_origin[0],
            self.search_origin[0] + self.search_size[0],
        ]
        d["search_y"] = [
            self.search_origin[1],
            self.search_origin[1] + self.search_size[1],
        ]
        d["search_z"] = [
            self.search_origin[2],
            self.search_origin[2] + self.search_size[2],
        ]
        for key, value in d.items():
            if isinstance(value, pathlib.Path):
                d[key] = str(value.absolute())
        # wrangle dtype conversion
        d["output_dtype"] = str(np.dtype(d["output_dtype"]))
        with open(file_name, "w") as fstream:
            json.dump(d, fstream, indent=4, cls=CustomJSONEncoder)

    def split_rotation_search(self, n: int) -> list[TMJob, ...]:
        """Split the search into sub_jobs by dividing the rotations. Sub jobs will
        obtain the key self.job_key + str(i) when looping over range(n).

        Parameters
        ----------
        n: int
            number of times to split the angular search

        Returns
        -------
        sub_jobs: list[TMJob, ...]
            a list of TMJobs that were split from self, the jobs are also assigned as
            the TMJob.sub_jobs attribute
        """
        if len(self.sub_jobs) > 0:
            raise TMJobError(
                "Could not further split this job as it already has subjobs assigned!"
            )

        sub_jobs = []
        for i in range(n):
            new_job = self.copy()
            new_job.start_slice = i
            new_job.steps_slice = n
            new_job.leader = self.job_key
            new_job.job_key = self.job_key + str(i)
            sub_jobs.append(new_job)

        self.sub_jobs = sub_jobs

        return self.sub_jobs

    def split_volume_search(self, split: tuple[int, int, int]) -> list[TMJob, ...]:
        """Split the search into sub_jobs by dividing into subvolumes. Final number of
        subvolumes is obtained by multiplying all the split together, e.g. (2, 2, 1)
        results in 4 subvolumes. Sub jobs will obtain the key self.job_key + str(i) when
        looping over range(n).

        The sub jobs search area of the full tomogram is defined by:
        new_job.search_origin and new_job.search_size.
        They are used when loading the search volume from the full tomogram.

        The attribute new_job.whole_start defines how the volume maps back to the score
        volume of the parent job (which can be a different size than the tomogram when
        the search is restricted along x, y or z).

        Finally, new_job.sub_start and new_job.sub_step, extract the score and angle map
        without the template overhang from the subvolume.

        If self.tomogram_mask is set, we will skip subjobs where all(mask <= 0).

        Parameters
        ----------
        split: tuple[int, int, int]
            tuple that defines how many times the search volume should be split into
            subvolumes along each axis

        Returns
        -------
        sub_jobs: list[TMJob, ...]
            a list of TMJobs that were split from self, the jobs are also assigned as
            the TMJob.sub_jobs attribute
        """
        if len(self.sub_jobs) > 0:
            raise TMJobError(
                "Could not further split this job as it already has subjobs assigned!"
            )

        search_size = self.search_size
        if self.tomogram_mask is not None:
            # This should have some positve values after the check in the __init__
            tomogram_mask = read_mrc(self.tomogram_mask)
        else:
            tomogram_mask = None
        # shape of template for overhang
        overhang = self.template_shape
        # use overhang//2 (+1 for odd sizes)
        overhang = tuple(sum(divmod(o, 2)) for o in overhang)

        x_splits = _determine_1D_fft_splits(search_size[0], split[0], overhang[0])
        y_splits = _determine_1D_fft_splits(search_size[1], split[1], overhang[1])
        z_splits = _determine_1D_fft_splits(search_size[2], split[2], overhang[2])

        sub_jobs = []
        for i, data_3D in enumerate(itt.product(x_splits, y_splits, z_splits)):
            # each data point for each dim is slice(left, right) of the search space
            # and slice(left,right) of the unique data point in the search space
            # Look at the comments in the new_job.attribute for the meaning of each
            # attribute

            search_origin = tuple(
                data_3D[d][0][0] + self.search_origin[d] for d in range(3)
            )
            search_size = tuple(dim_data[0][1] - dim_data[0][0] for dim_data in data_3D)
            whole_start = tuple(dim_data[1][0] for dim_data in data_3D)
            sub_start = tuple(dim_data[1][0] - dim_data[0][0] for dim_data in data_3D)
            sub_step = tuple(dim_data[1][1] - dim_data[1][0] for dim_data in data_3D)

            # check if this contains any of the unique data points are where
            # tomo_mask > 0
            if tomogram_mask is not None:
                slices = [
                    slice(origin, origin + step)
                    for origin, step in zip(whole_start, sub_step)
                ]
                if np.all(tomogram_mask[*slices] <= 0):
                    # No non-masked unique data-points, skipping
                    continue
            new_job = self.copy()
            new_job.leader = self.job_key
            new_job.job_key = self.job_key + str(i)

            # search origin with respect to the complete tomogram
            new_job.search_origin = search_origin
            # search size TODO: should be combined with the origin into slices
            new_job.search_size = search_size

            # whole start is the start of the unique data within the complete searched
            # array
            new_job.whole_start = whole_start
            # sub_start is where the unique data starts inside the split array
            new_job.sub_start = sub_start
            # sub_step is the step of unique data inside the split array.
            # TODO: should be slices instead
            new_job.sub_step = sub_step
            sub_jobs.append(new_job)

        self.sub_jobs = sub_jobs

        return self.sub_jobs

    def merge_sub_jobs(
        self, stats: list[dict, ...] | None = None
    ) -> tuple[npt.NDArray[float], npt.NDArray[float]]:
        """Merge the sub jobs present in self.sub_jobs together to create the final
        output score and angle maps.

        Parameters
        ----------
        stats: Optional[list[dict, ...]], default None
            optional list of sub job statistics to merge together

        Returns
        -------
        output: tuple[npt.NDArray[float], npt.NDArray[float]]
            the merged score and angle maps from the subjobs
        """
        if len(self.sub_jobs) == 0:
            # read the volumes, remove them and return them
            score_file, angle_file = (
                self.output_dir.joinpath(f"{self.tomo_id}_scores_{self.job_key}.mrc"),
                self.output_dir.joinpath(f"{self.tomo_id}_angles_{self.job_key}.mrc"),
            )
            result = (read_mrc(score_file), read_mrc(angle_file))
            (score_file.unlink(), angle_file.unlink())
            return result

        if stats is not None:
            search_space = sum([s["search_space"] for s in stats])
            variance = sum([s["variance"] for s in stats]) / len(stats)
            self.job_stats = {
                "search_space": search_space,
                "variance": variance,
                "std": np.sqrt(variance),
            }

        is_subvolume_split = np.all(
            np.array([x.start_slice for x in self.sub_jobs]) == 0
        )

        score_volumes, angle_volumes = [], []
        for x in self.sub_jobs:
            result = x.merge_sub_jobs()
            score_volumes.append(result[0])
            angle_volumes.append(result[1])

        if not is_subvolume_split:
            scores, angles = (
                np.zeros_like(score_volumes[0]) - 1.0,
                np.zeros_like(angle_volumes[0]) - 1.0,
            )
            for s, a in zip(score_volumes, angle_volumes):
                angles = np.where(s > scores, a, angles)
                # prevents race condition due to slicing
                angles = np.where(s == scores, np.minimum(a, angles), angles)
                scores = np.where(s > scores, s, scores)
        else:
            scores, angles = (
                np.zeros(self.search_size, dtype=np.float32),
                np.zeros(self.search_size, dtype=np.float32),
            )
            for job, s, a in zip(self.sub_jobs, score_volumes, angle_volumes):
                sub_scores = s[
                    job.sub_start[0] : job.sub_start[0] + job.sub_step[0],
                    job.sub_start[1] : job.sub_start[1] + job.sub_step[1],
                    job.sub_start[2] : job.sub_start[2] + job.sub_step[2],
                ]
                sub_angles = a[
                    job.sub_start[0] : job.sub_start[0] + job.sub_step[0],
                    job.sub_start[1] : job.sub_start[1] + job.sub_step[1],
                    job.sub_start[2] : job.sub_start[2] + job.sub_step[2],
                ]
                # Then the corrected sub part needs to be placed back into the full
                # volume
                scores[
                    job.whole_start[0] : job.whole_start[0] + sub_scores.shape[0],
                    job.whole_start[1] : job.whole_start[1] + sub_scores.shape[1],
                    job.whole_start[2] : job.whole_start[2] + sub_scores.shape[2],
                ] = sub_scores
                angles[
                    job.whole_start[0] : job.whole_start[0] + sub_scores.shape[0],
                    job.whole_start[1] : job.whole_start[1] + sub_scores.shape[1],
                    job.whole_start[2] : job.whole_start[2] + sub_scores.shape[2],
                ] = sub_angles
        return scores.astype(self.output_dtype), angles

    def start_job(
        self, gpu_id: int, return_volumes: bool = False
    ) -> tuple[npt.NDArray[float], npt.NDArray[float]] | dict:
        """Run this template matching job on the specified GPU. Search statistics of the
        job will always be assigned to the self.job_stats.

        Parameters
        ----------
        gpu_id: int
            index of the GPU to run the job on
        return_volumes: bool, default False
            False (default) does not return volumes but instead writes them to disk, set
            to True to instead directly return the score and angle volumes

        Returns
        -------
        output: Union[tuple[npt.NDArray[float], npt.NDArray[float]], dict]
            when volumes are returned the output consists of two numpy arrays (score and
            angle map), when no volumes are returned the output consists of a dictionary
            with search statistics
        """
        tomo = read_mrc(self.tomogram)
        fast_tomo = np.zeros(
            tuple([next_fast_len(s, real=True) for s in tomo.shape]),
            dtype=np.float32,
        )
        fast_tomo[: tomo.shape[0], : tomo.shape[1], : tomo.shape[2]] = tomo

        # load template and mask
        template, mask = (read_mrc(self.template), read_mrc(self.mask))

        # init tomogram and template weighting
        tomo_filter, template_wedge = 1, 1
        # first generate bandpass filters
        if not (self.low_pass is None and self.high_pass is None):
            tomo_filter *= create_gaussian_band_pass(
                fast_tomo.shape, self.voxel_size, self.low_pass, self.high_pass
            ).astype(np.float32)
            template_wedge *= create_gaussian_band_pass(
                self.template_shape, self.voxel_size, self.low_pass, self.high_pass
            ).astype(np.float32)

        # then multiply with optional whitening filters
        if self.whiten_spectrum:
            tomo_filter *= profile_to_weighting(
                np.load(self.whitening_filter), fast_tomo.shape
            ).astype(np.float32)
            template_wedge *= profile_to_weighting(
                np.load(self.whitening_filter), self.template_shape
            ).astype(np.float32)

        # create wedge filters
        if (
            self.ts_metadata.per_tilt_weighting
            and self.ts_metadata.defocus_handedness != 0
        ):
            # adjust ctf parameters for this specific patch in the tomogram
            full_tomo_center = np.array(self.tomo_shape) / 2
            patch_center = np.array(self.search_origin) + np.array(self.search_size) / 2
            relative_patch_center_angstrom = (
                patch_center - full_tomo_center
            ) * self.voxel_size
            defocus_offsets = get_defocus_offsets(
                relative_patch_center_angstrom[0],  # x-coordinate
                relative_patch_center_angstrom[2],  # z-coordinate
                self.ts_metadata,
            )
            # TODO: make sure this doesn't lead to weird race conditions
            for ctf, defocus_shift in zip(self.ts_metadata.ctf_data, defocus_offsets):
                ctf.defocus = ctf.defocus + defocus_shift * 1e-10
            logging.debug(
                "Patch center (nr. of voxels): "
                f"{np.array_str(relative_patch_center_angstrom, precision=2)}"
            )
            logging.debug(
                "Defocus values (um): "
                f"{[round(ctf.defocus * 1e6, 2) for ctf in self.ts_metadata.ctf_data]}",
            )

        # for the tomogram a binary wedge is generated to explicitly set the missing
        # wedge region to 0
        tomo_filter *= create_wedge(
            fast_tomo.shape,
            self.ts_metadata,
            self.voxel_size,
            cut_off_radius=1.0,
            per_tilt_weighting=False,
        ).astype(np.float32)
        # for the template a binary or per-tilt-weighted wedge is generated
        # depending on the options
        template_wedge *= create_wedge(
            self.template_shape,
            self.ts_metadata,
            self.voxel_size,
            cut_off_radius=1.0,
        ).astype(np.float32)

        if logging.DEBUG >= logging.root.level:
            write_mrc(
                self.output_dir.joinpath("template_psf.mrc"),
                template_wedge,
                self.voxel_size,
            )
            write_mrc(
                self.output_dir.joinpath("template_convolved.mrc"),
                irfftn(rfftn(template) * template_wedge, s=template.shape),
                self.voxel_size,
            )

        # next fast fft len
        logging.debug(
            "Next fast fft shape: "
            f"{tuple([next_fast_len(s, real=True) for s in self.search_size])}"
        )
        search_volume = np.zeros(
            tuple([next_fast_len(s, real=True) for s in self.search_size]),
            dtype=np.float32,
        )

        # apply optional filters to tomogram and slice it down to subvolume
        # TODO: we might want to rewrite this to only do this and the tomo loading
        # once in a volume split
        fast_tomo = np.real(irfftn(rfftn(fast_tomo) * tomo_filter, s=fast_tomo.shape))
        search_volume[
            : self.search_size[0], : self.search_size[1], : self.search_size[2]
        ] = np.ascontiguousarray(
            fast_tomo[
                self.search_origin[0] : self.search_origin[0] + self.search_size[0],
                self.search_origin[1] : self.search_origin[1] + self.search_size[1],
                self.search_origin[2] : self.search_origin[2] + self.search_size[2],
            ]
        )

        # load rotation search
        angle_ids = list(range(self.start_slice, self.n_rotations, self.steps_slice))
        angle_list = get_angle_list(
            self.rotation_file,
            sort_angles=version.parse(self.pytom_tm_version_number)
            > version.parse("0.3.0"),
            symmetry=self.rotational_symmetry,
        )

        angle_list = angle_list[
            slice(self.start_slice, self.n_rotations, self.steps_slice)
        ]

        # slices for relevant part for job statistics
        search_volume_roi = (
            slice(self.sub_start[0], self.sub_start[0] + self.sub_step[0]),
            slice(self.sub_start[1], self.sub_start[1] + self.sub_step[1]),
            slice(self.sub_start[2], self.sub_start[2] + self.sub_step[2]),
        )

        tm = TemplateMatchingGPU(
            job_id=self.job_key,
            device_id=gpu_id,
            volume=search_volume,
            template=template,
            mask=mask,
            angle_list=angle_list,
            angle_ids=angle_ids,
            mask_is_spherical=self.mask_is_spherical,
            wedge=template_wedge,
            stats_roi=search_volume_roi,
            noise_correction=self.random_phase_correction,
            rng_seed=self.rng_seed,
        )
        results = tm.run()
        score_volume = results[0][
            : self.search_size[0], : self.search_size[1], : self.search_size[2]
        ]
        angle_volume = results[1][
            : self.search_size[0], : self.search_size[1], : self.search_size[2]
        ]
        self.job_stats = results[2]

        del tm  # delete the template matching plan

        # cast to correct dtype
        score_volume = score_volume.astype(self.output_dtype)
        angle_volume = angle_volume

        if return_volumes:
            return score_volume, angle_volume
        else:  # otherwise write them out with job_key
            write_mrc(
                self.output_dir.joinpath(f"{self.tomo_id}_scores_{self.job_key}.mrc"),
                score_volume,
                self.voxel_size,
            )
            write_mrc(
                self.output_dir.joinpath(f"{self.tomo_id}_angles_{self.job_key}.mrc"),
                angle_volume,
                self.voxel_size,
            )
            return self.job_stats



"""
Functions in this file are cpu/gpu agnostic.
"""

import numpy.typing as npt
import cupy.typing as cpt


def mean_under_mask(
    data: npt.NDArray[float] | cpt.NDArray[float],
    mask: npt.NDArray[float] | cpt.NDArray[float],
    mask_weight: float | None = None,
) -> float | cpt.NDArray[float]:
    """Calculate mean of array in the mask region.

    data and mask can be cupy or numpy arrays.

    Parameters
    ----------
    data: Union[npt.NDArray[float], cpt.NDArray[float]]
        input array
    mask: Union[npt.NDArray[float], cpt.NDArray[float]]
        input mask, same dimensions as data
    mask_weight: Optional[float], default None
        optional weight of mask, if not provided mask.sum() is used to determine weight

    Returns
    -------
    output: Union[float, cpt.NDArray[float]]
        mean of data in the region of the mask
    """
    output = (data * mask).sum() / (
        mask_weight if mask_weight is not None else mask.sum()
    )
    return output


def std_under_mask(
    data: npt.NDArray[float] | cpt.NDArray[float],
    mask: npt.NDArray[float] | cpt.NDArray[float],
    mean: float,
    mask_weight: float | None = None,
) -> float | cpt.NDArray[float]:
    """Calculate standard deviation of array in the mask region. Uses mean_under_mask()
    to calculate the mean of data**2 within the mask.

    data and mask can be cupy or numpy arrays.

    Parameters
    ----------
    data: Union[npt.NDArray[float], cpt.NDArray[float]]
        input array
    mask: Union[npt.NDArray[float], cpt.NDArray[float]]
        input mask, same dimensions as data
    mean: float
        mean of array in masked region
    mask_weight: Optional[float], default None
        optional weight of mask, if not provided mask.sum() is used to determine weight

    Returns
    -------
    output: Union[float, cpt.NDArray[float]]
        standard deviation of data in the region of the mask
    """
    output = (mean_under_mask(data**2, mask, mask_weight=mask_weight) - mean**2) ** 0.5
    return output


def normalise(
    data: npt.NDArray[float] | cpt.NDArray[float],
    mask: npt.NDArray[float] | cpt.NDArray[float] | None = None,
    mask_weight: float | None = None,
) -> npt.NDArray[float] | cpt.NDArray[float]:
    """Normalise array by subtracting mean and dividing by standard deviation. If a mask
    is provided the array is normalised with the mean and std calculated within the
    mask.

    data and mask can be cupy or numpy arrays.

    Parameters
    ----------
    data: Union[npt.NDArray[float], cpt.NDArray[float]]
        input array to normalise
    mask: Optional[Union[npt.NDArray[float], cpt.NDArray[float]]], default None
        optional mask to normalise with mean and std in masked region
    mask_weight: Optional[float], default None
        optional float specifying mask weight, if not provided mask.sum() is used

    Returns
    -------
    output: Union[npt.NDArray[float], cpt.NDArray[float]]
        normalised array
    """
    if mask is None:
        mean, std = data.mean(), data.std()
    else:
        mean = mean_under_mask(data, mask, mask_weight=mask_weight)
        std = std_under_mask(data, mask, mean, mask_weight=mask_weight)
    output = (data - mean) / std
    return output


def normalised_cross_correlation(
    data1: npt.NDArray[float] | cpt.NDArray[float],
    data2: npt.NDArray[float] | cpt.NDArray[float],
) -> float | cpt.NDArray[float]:
    """Calculate normalised cross correlation between two arrays.

    data1, and data2 can be cupy or numpy arrays.

    Parameters
    ----------
    data1: Union[npt.NDArray[float], cpt.NDArray[float]]
        first array for correlation
    data2: Union[npt.NDArray[float], cpt.NDArray[float]]
        second array for correlation

    Returns
    -------
    output: Union[float, cpt.NDArray[float]]
        normalised cross correlation between the arrays
    """
    output = (normalise(data1) * normalise(data2)).sum() / data1.size
    return output



import cupy as cp
import cupy.typing as cpt
import numpy.typing as npt
import voltools as vt
import gc
from cupyx.scipy.fft import rfftn, irfftn
from tqdm import tqdm
from pytom_tm.correlation import normalise
from pytom_tm.template import phase_randomize_template


class TemplateMatchingPlan:
    def __init__(
        self,
        volume: npt.NDArray[float],
        template: npt.NDArray[float],
        mask: npt.NDArray[float],
        device_id: int,
        wedge: npt.NDArray[float] | None = None,
        phase_randomized_template: npt.NDArray[float] | None = None,
    ):
        """Initialize a template matching plan. All the necessary cupy arrays will be
        allocated on the GPU.

        Parameters
        ----------
        volume: npt.NDArray[float]
            3D numpy array representing the search tomogram
        template: npt.NDArray[float]
            3D numpy array representing the template for the search, a square box of
            size sx
        mask: npt.NDArray[float]
            3D numpy array representing the mask for the search, same dimensions as
            template
        device_id: int
            GPU device id to load arrays on
        wedge: Optional[npt.NDArray[float]], default None
            3D numpy array that contains the Fourier space weighting for the template,
            it should be in Fourier reduced form, with dimensions (sx, sx, sx // 2 + 1)
        phase_randomized_template: Optional[npt.NDArray[float]], default None
            initialize the plan with a phase randomized version of the template for
            noise correction
        """
        # Search volume + and fft transform plan for the volume
        volume_shape = volume.shape
        cp_vol = cp.asarray(volume, dtype=cp.float32, order="C")
        self.volume_rft_conj = rfftn(cp_vol).conj()
        self.volume_sq_rft_conj = rfftn(cp_vol**2).conj()
        # Explicit fft plan is no longer necessary as cupy generates a plan behind the
        # scene which leads to comparable timings

        # Array for storing local standard deviations
        self.std_volume = cp.zeros(volume_shape, dtype=cp.float32)

        # Data for the mask
        self.mask = cp.asarray(mask, dtype=cp.float32, order="C")
        self.mask_texture = vt.StaticVolume(
            self.mask, interpolation="filt_bspline", device=f"gpu:{device_id}"
        )
        self.mask_padded = cp.zeros(volume_shape, dtype=cp.float32)
        self.mask_weight = self.mask.sum()  # weight of the mask

        # Init template data
        self.template = cp.asarray(template, dtype=cp.float32, order="C")
        self.template_texture = vt.StaticVolume(
            self.template, interpolation="filt_bspline", device=f"gpu:{device_id}"
        )
        self.template_padded = cp.zeros(volume_shape, dtype=cp.float32)

        # fourier binary wedge weight for the template
        self.wedge = (
            cp.asarray(wedge, order="C", dtype=cp.float32)
            if wedge is not None
            else None
        )

        # Initialize result volumes
        self.ccc_map = cp.zeros(volume_shape, dtype=cp.float32)
        self.scores = cp.ones(volume_shape, dtype=cp.float32) * -1000
        self.angles = cp.ones(volume_shape, dtype=cp.float32) * -1000

        self.random_phase_template_texture = None
        self.noise_scores = None
        if phase_randomized_template is not None:
            self.random_phase_template_texture = vt.StaticVolume(
                cp.asarray(phase_randomized_template, dtype=cp.float32, order="C"),
                interpolation="filt_bspline",
                device=f"gpu:{device_id}",
            )
            self.noise_scores = cp.ones(volume_shape, dtype=cp.float32) * -1000

        # wait for stream to complete the work
        cp.cuda.stream.get_current_stream().synchronize()

    def clean(self) -> None:
        """Remove all stored cupy arrays from the GPU's memory pool."""
        gpu_memory_pool = cp.get_default_memory_pool()
        del (
            self.volume_rft_conj,
            self.volume_sq_rft_conj,
            self.mask,
            self.mask_texture,
            self.mask_padded,
            self.template,
            self.template_texture,
            self.template_padded,
            self.wedge,
            self.ccc_map,
            self.scores,
            self.angles,
            self.std_volume,
        )
        gc.collect()
        gpu_memory_pool.free_all_blocks()


class TemplateMatchingGPU:
    def __init__(
        self,
        job_id: str,
        device_id: int,
        volume: npt.NDArray[float],
        template: npt.NDArray[float],
        mask: npt.NDArray[float],
        angle_list: list[tuple[float, float, float]],
        angle_ids: list[int],
        mask_is_spherical: bool = True,
        wedge: npt.NDArray[float] | None = None,
        stats_roi: tuple[slice, slice, slice] | None = None,
        noise_correction: bool = False,
        rng_seed: int = 321,
    ):
        """Initialize a template matching run.

        For other great implementations see:
        - STOPGAP: https://github.com/wan-lab-vanderbilt/STOPGAP
        - pyTME: https://github.com/KosinskiLab/pyTME

        The precalculation of conjugated FTs of the tomo was (AFAIK) introduced
        by STOPGAP! Also, they introduced simultaneous matching with a phase randomized
        version of the template. https://doi.org/10.1107/S205979832400295X

        Parameters
        ----------
        job_id: str
            string for job identification
        device_id: int
            GPU device id to run the job on
        volume: npt.NDArray[float]
            3D numpy array of tomogram
        template: npt.NDArray[float]
            3D numpy array of template, square box of size sx
        mask: npt.NDArray[float]
            3D numpy array with mask, same box size as template
        angle_list: list[tuple[float, float, float]]
            list of tuples with 3 floats representing Euler angle rotations
        angle_ids: list[int]
            list of indices for angle_list to actually search, this can be a subset of
            the full list
        mask_is_spherical: bool, default True
            True (default) if mask is spherical, set to False for non-spherical mask
            which increases computation time
        wedge: Optional[npt.NDArray[float]], default None
            3D numpy array that contains the Fourier space weighting for the template,
            it should be in Fourier reduced form, with dimensions (sx, sx, sx // 2 + 1)
        stats_roi: Optional[tuple[slice, slice, slice]], default None
            region of interest to calculate statistics on the search volume, default
            will just take the full search volume
        noise_correction: bool, default False
            initialize template matching with a phase randomized version of the template
            that is used to subtract background noise from the score map; expense is
            more gpu memory and computation time
        rng_seed: int, default 321
            seed for rng in phase randomization
        """
        cp.cuda.Device(device_id).use()

        self.job_id = job_id
        self.device_id = device_id
        self.active = True
        self.completed = False
        self.mask_is_spherical = mask_is_spherical  # whether mask is spherical
        self.angle_list = angle_list
        self.angle_ids = angle_ids
        self.stats = {"search_space": 0, "variance": 0.0, "std": 0.0}
        if stats_roi is None:
            self.stats_roi = (slice(None), slice(None), slice(None))
        else:
            self.stats_roi = stats_roi
        self.noise_correction = noise_correction

        # create a 'random noise' version of the template
        shuffled_template = (
            phase_randomize_template(template, rng_seed) if noise_correction else None
        )

        self.plan = TemplateMatchingPlan(
            volume,
            template,
            mask,
            device_id,
            wedge=wedge,
            phase_randomized_template=shuffled_template,
        )

    def run(self) -> tuple[npt.NDArray[float], npt.NDArray[float], dict]:
        """Run the template matching job.

        Returns
        -------
        results: tuple[npt.NDArray[float], npt.NDArray[float], dict]
            results is a tuple with tree elements:
                - score_map with the locally normalised maximum score over all the
                angles searched;
                    a 3D numpy array with same size as search volume
                - angle_map with an index into the angle list corresponding to the
                maximum of the correlation scores;
                    a 3D numpy array with same size as search volume
                - a dictionary with three floats of search statistics;
                    'search_space', 'variance', and 'std'
        """
        print(f"Progress job_{self.job_id} on device {self.device_id:d}:")

        # Size x template (sxz) and center x template (cxt)
        sxt, syt, szt = self.plan.template.shape
        cxt, cyt, czt = sxt // 2, syt // 2, szt // 2
        mx, my, mz = sxt % 2, syt % 2, szt % 2  # odd or even

        # Size x volume (sxv) and center x volume (xcv)
        sxv, syv, szv = self.plan.template_padded.shape
        cxv, cyv, czv = sxv // 2, syv // 2, szv // 2

        # create slice for padding
        pad_index = (
            slice(cxv - cxt, cxv + cxt + mx),
            slice(cyv - cyt, cyv + cyt + my),
            slice(czv - czt, czv + czt + mz),
        )

        # calculate roi mask
        shift = cp.floor(cp.array(self.plan.scores.shape) / 2).astype(int) + 1
        roi_mask = cp.zeros(self.plan.scores.shape, dtype=bool)
        roi_mask[self.stats_roi] = True
        roi_mask = cp.flip(cp.roll(roi_mask, -shift, (0, 1, 2)))
        roi_size = self.plan.scores[roi_mask].size

        if self.mask_is_spherical:  # Then we only need to calculate std volume once
            self.plan.mask_padded[pad_index] = self.plan.mask
            self.plan.std_volume = (
                std_under_mask_convolution(
                    self.plan.volume_rft_conj,
                    self.plan.volume_sq_rft_conj,
                    self.plan.mask_padded,
                    self.plan.mask_weight,
                )
                * self.plan.mask_weight
            )

        # Track iterations with a tqdm progress bar
        for i in tqdm(range(len(self.angle_ids))):
            # tqdm cannot loop over zipped lists, so need to do it like this
            angle_id, rotation = self.angle_ids[i], self.angle_list[i]

            if not self.mask_is_spherical:
                self.plan.mask_texture.transform(
                    rotation=(rotation[0], rotation[1], rotation[2]),
                    rotation_order="rzxz",
                    output=self.plan.mask,
                    rotation_units="rad",
                )
                self.plan.mask_padded[pad_index] = self.plan.mask
                # Std volume needs to be recalculated for every rotation of the mask,
                # expensive step
                self.plan.std_volume = (
                    std_under_mask_convolution(
                        self.plan.volume_rft_conj,
                        self.plan.volume_sq_rft_conj,
                        self.plan.mask_padded,
                        self.plan.mask_weight,
                    )
                    * self.plan.mask_weight
                )

            # Rotate template
            self.plan.template_texture.transform(
                rotation=(rotation[0], rotation[1], rotation[2]),
                rotation_order="rzxz",
                output=self.plan.template,
                rotation_units="rad",
            )

            self.correlate(pad_index)

            # Update the scores and angle_lists
            update_results_kernel(
                self.plan.scores,
                self.plan.ccc_map,
                angle_id,
                self.plan.scores,
                self.plan.angles,
            )

            self.stats["variance"] += (
                square_sum_kernel(self.plan.ccc_map * roi_mask) / roi_size
            )

            if self.noise_correction:
                # Rotate noise template texture into template
                self.plan.random_phase_template_texture.transform(
                    rotation=(rotation[0], rotation[1], rotation[2]),
                    rotation_order="rzxz",
                    output=self.plan.template,
                    rotation_units="rad",
                )

                self.correlate(pad_index)

                # update noise scores results
                update_noise_template_results_kernel(
                    self.plan.noise_scores,
                    self.plan.ccc_map,
                    self.plan.noise_scores,
                )

        # do the noise correction on the scores map: substract the noise scores first,
        # and then add the noise mean to ensure stats are consistent
        if self.noise_correction:
            self.plan.scores = (
                self.plan.scores - self.plan.noise_scores
            ) + self.plan.noise_scores.mean()

        # Get correct orientation back!
        # Use same method as William Wan's STOPGAP
        # (https://doi.org/10.1107/S205979832400295X): the search volume is Fourier
        # transformed and conjugated before the iterations this means the eventual
        # score map needs to be flipped back. The map is also rolled due to the ftshift
        # effect of a Fourier space correlation function.
        self.plan.scores = cp.roll(cp.flip(self.plan.scores), shift, axis=(0, 1, 2))
        self.plan.angles = cp.roll(cp.flip(self.plan.angles), shift, axis=(0, 1, 2))

        self.stats["search_space"] = int(roi_size * len(self.angle_ids))
        self.stats["variance"] = float(self.stats["variance"] / len(self.angle_ids))
        self.stats["std"] = float(cp.sqrt(self.stats["variance"]))

        # package results back to the CPU
        results = (self.plan.scores.get(), self.plan.angles.get(), self.stats)

        # clear all the used gpu memory
        self.plan.clean()

        return results

    def correlate(self, padding_index: tuple[slice, slice, slice]):
        """Correlate template and tomogram.

        Parameters
        ----------
        padding_index: tuple[slice, slice, slice]
            Location to pad template after weighting and normalization
        """
        if self.plan.wedge is not None:
            # Add wedge to the template after rotating
            self.plan.template = irfftn(
                rfftn(self.plan.template) * self.plan.wedge, s=self.plan.template.shape
            )

        # Normalize and mask template
        norm = normalise(
            self.plan.template, self.plan.mask, mask_weight=self.plan.mask_weight
        )
        self.plan.template = norm * self.plan.mask

        # Paste in center
        self.plan.template_padded[padding_index] = self.plan.template

        # Fast local correlation function between volume and template, norm is the
        # standard deviation at each point in the volume in the masked area
        self.plan.ccc_map = (
            irfftn(
                self.plan.volume_rft_conj * rfftn(self.plan.template_padded),
                s=self.plan.template_padded.shape,
            )
            / self.plan.std_volume
        )


def std_under_mask_convolution(
    volume_rft_conj: cpt.NDArray[float],
    volume_sq_rft_conj: cpt.NDArray[float],
    padded_mask: cpt.NDArray[float],
    mask_weight: float,
) -> cpt.NDArray[float]:
    """Calculate the local standard deviation under the mask for each position in the
    volume. Calculation is done in Fourier space as this is a convolution between volume
    and mask.

    Parameters
    ----------
    volume_rft_conj: cpt.NDArray[float]
        complex conjugate of the rft of the search volume
    volume_sq_rft_conj: cpt.NDArray[float]
        complex conjugate of the rft of the squared search volume
    padded_mask: cpt.NDArray[float]
        template mask that has been padded to dimensions of volume
    mask_weight: float
        weight of the mask, usually calculated as mask.sum()

    Returns
    -------
    std_v: cpt.NDArray[float]
        array with local standard deviations in volume
    """
    padded_mask_rft = rfftn(padded_mask)
    std_v = (
        irfftn(volume_sq_rft_conj * padded_mask_rft, s=padded_mask.shape) / mask_weight
        - (irfftn(volume_rft_conj * padded_mask_rft, s=padded_mask.shape) / mask_weight)
        ** 2
    )
    std_v[std_v <= cp.float32(1e-18)] = (
        1  # prevent potential sqrt of negative value and division by zero
    )
    std_v = cp.sqrt(std_v)
    return std_v


"""Update scores and angles if a new maximum is found."""
update_results_kernel = cp.ElementwiseKernel(
    "float32 scores, float32 ccc_map, float32 angles",
    "float32 scores_out, float32 angles_out",
    "if (scores < ccc_map) {scores_out = ccc_map; angles_out = angles;}",
    "update_results",
)


"""Update scores for noise template"""
update_noise_template_results_kernel = cp.ElementwiseKernel(
    "float32 scores, float32 ccc_map",
    "float32 scores_out",
    "if (scores < ccc_map) {scores_out = ccc_map;}",
    "update_noise_template_results",
)


"""Calculate the sum of squares in a volume. Mean is assumed to be 0 which makes this 
operation a lot faster."""
square_sum_kernel = cp.ReductionKernel(
    "T x",  # input params
    "T y",  # output params
    "x * x",  # pre-processing expression
    "a + b",  # reduction operation
    "y = a",  # post-reduction output processing
    "0",  # identity value
    "variance",  # kernel name
)



import unittest
from pytom_tm.dataclass import CtfData, TiltSeriesMetaData


class TestCtfDataclass(unittest.TestCase):
    def test_replace(self):
        # test that replace exists and can be used
        temp = CtfData(
            defocus=1, amplitude_contrast=0.1, voltage=300, spherical_aberration=1e-6
        )
        temp2 = temp.replace(defocus=2)
        self.assertNotEqual(temp.defocus, temp2.defocus)
        self.assertEqual(temp.amplitude_contrast, temp2.amplitude_contrast)
        self.assertEqual(temp.voltage, temp2.voltage)
        self.assertEqual(temp.spherical_aberration, temp2.spherical_aberration)


class TestTiltSeriesDataclass(unittest.TestCase):
    def test_sanity_errors(self):
        with self.assertRaisesRegex(ValueError, "at least 2 tilt angles"):
            _ = TiltSeriesMetaData(tilt_angles=[1.0])
        with self.assertRaisesRegex(ValueError, "at least 2 tilt angles"):
            _ = TiltSeriesMetaData(tilt_angles=1.0)

        a = CtfData(
            defocus=1, amplitude_contrast=0.1, voltage=300, spherical_aberration=1e-6
        )
        with self.assertRaisesRegex(ValueError, "a single CtfData or the same number"):
            _ = TiltSeriesMetaData(tilt_angles=[1, 2, 3], ctf_data=[a, a])

        with self.assertRaisesRegex(ValueError, "invalid defocus handedness"):
            _ = TiltSeriesMetaData(tilt_angles=[1, 2, 3], defocus_handedness=0.5)

    def test_sanity_on_replace(self):
        # test that the sanity checks happen when replacing values
        temp = TiltSeriesMetaData(tilt_angles=[1.0, 2.0], dose_accumulation=[1, 2])
        with self.assertRaisesRegex(
            ValueError, r"same number of doses as tilt angles \(3\)"
        ):
            _ = temp.replace(tilt_angles=[1, 2, 3])

    def test_ctf_expansion(self):
        a = CtfData(
            defocus=1, amplitude_contrast=0.1, voltage=300, spherical_aberration=1e-6
        )
        ts_metadata = TiltSeriesMetaData(tilt_angles=[1, 2, 3], ctf_data=[a])
        self.assertEqual(len(ts_metadata.tilt_angles), len(ts_metadata.ctf_data))
        for ctf in ts_metadata.ctf_data:
            self.assertIs(ctf, a)



import numpy.typing as npt
import numpy as np
import voltools as vt
import logging
from scipy.ndimage import center_of_mass, zoom
from scipy.fft import rfftn, irfftn
from pytom_tm.weights import (
    create_gaussian_low_pass,
    radial_reduced_grid,
)


def generate_template_from_map(
    input_map: npt.NDArray[float],
    input_spacing: float,
    output_spacing: float,
    center: bool = False,
    filter_to_resolution: float | None = None,
    output_box_size: int | None = None,
) -> npt.NDArray[float]:
    """Generate a template from a density map.

    Parameters
    ----------
    input_map: npt.NDArray[float]
        3D density to use for generating the template, if box is not square it will be
        padded to square
    input_spacing: float
        voxel size of input map (in A)
    output_spacing: float
        voxel size of output map (in A) the ratio of input to output will be used for
        downsampling
    center: bool, default False
        set to True to center the template in the box by calculating the center of mass
    filter_to_resolution: Optional[float], default None
        low-pass filter resolution to apply to template, if not provided will be set to
        2 * output_spacing
    output_box_size:  Optional[int], default None
        final box size of template
    display_filter: bool, default False
        flag to display a plot of the filter applied to the template

    Returns
    -------
    template: npt.NDArray[float]
        processed template in the specified output box size, box will be square
    """
    # make the map a box with equal dimensions
    if len(set(input_map.shape)) != 1:
        diff = [max(input_map.shape) - s for s in input_map.shape]
        input_map = np.pad(
            input_map,
            tuple([(d // 2, d // 2 + d % 2) for d in diff]),
            mode="constant",
            constant_values=0,
        )

    if filter_to_resolution is None:
        # Set to nyquist resolution
        filter_to_resolution = 2 * output_spacing
    elif filter_to_resolution < (2 * output_spacing):
        warning_text = (
            f"Filter resolution is too low,"
            f" setting to {2 * output_spacing}A (2 * output voxel size)"
        )
        logging.warning(warning_text)
        filter_to_resolution = 2 * output_spacing

    if center:
        volume_center = np.divide(np.subtract(input_map.shape, 1), 2, dtype=np.float32)
        # square input to make values positive for center of mass
        input_center_of_mass = center_of_mass(input_map**2)
        shift = np.subtract(volume_center, input_center_of_mass)
        input_map = vt.transform(input_map, translation=shift, device="cpu")

        logging.debug(
            f"center of mass, before was "
            f"{np.round(input_center_of_mass, 2)} "
            f"and after {np.round(center_of_mass(input_map**2), 2)}"
        )

    # extend volume to the desired output size before applying convolutions!
    if output_box_size is not None:
        logging.debug(
            f"size check {output_box_size} > "
            f"{(input_map.shape[0] * input_spacing) // output_spacing}"
        )
        if output_box_size > (input_map.shape[0] * input_spacing) // output_spacing:
            pad = (
                int(output_box_size * (output_spacing / input_spacing))
                - input_map.shape[0]
            )
            logging.debug(f"pad with this number of zeros: {pad}")
            input_map = np.pad(
                input_map,
                (pad // 2, pad // 2 + pad % 2),
                mode="constant",
                constant_values=0,
            )
        elif output_box_size < (input_map.shape[0] * input_spacing) // output_spacing:
            logging.warning(
                "Could not set specified box size as the map would need to be cut and "
                "this might result in loss of information of the structure. Please "
                "decrease the box size of the map by hand (e.g. chimera)"
            )

    # create low pass filter
    lpf = create_gaussian_low_pass(
        input_map.shape, input_spacing, filter_to_resolution
    ).astype(np.float32)

    logging.info("Convoluting volume with filter and then downsampling.")
    return zoom(
        irfftn(rfftn(input_map) * lpf, s=input_map.shape),
        input_spacing / output_spacing,
    )


def phase_randomize_template(
    template: npt.NDArray[float],
    seed: int = 321,
):
    """Create a version of the template that has its phases randomly
    permuted in Fourier space.

    Parameters
    ----------
    template: npt.NDArray[float]
        input structure
    seed: int, default 321
        seed for random number generator for phase permutation

    Returns
    -------
    result: npt.NDArray[float]
        phase randomized version of the template
    """
    ft = rfftn(template)
    amplitude = np.abs(ft)

    # permute the phases in flattened version of the array
    phase = np.angle(ft).flatten()
    grid = np.fft.ifftshift(radial_reduced_grid(template.shape), axes=(0, 1)).flatten()
    relevant_freqs = grid <= 1  # permute only up to Nyquist
    noise = np.zeros_like(phase)
    rng = np.random.default_rng(seed)
    noise[relevant_freqs] = rng.permutation(phase[relevant_freqs])

    # construct the new template
    noise = np.reshape(noise, amplitude.shape)
    result = irfftn(amplitude * np.exp(1j * noise), s=template.shape)
    return result



import pathlib
import os
import contextlib
import numpy as np
import pandas as pd
import starfile
from pytom_tm.io import read_defocus_file, read_dose_file, read_tlt_file
from pytom_tm.dataclass import CtfData

# Dose and ctf params for tomo_104
cs = 2.7
amp = 0.08
vol = 200
defocus_data = read_defocus_file(
    pathlib.Path(__file__).parent.joinpath("Data").joinpath("test_imod.defocus")
)
CTF_PARAMS = []
for d in defocus_data:
    CTF_PARAMS.append(
        CtfData(
            defocus=d,
            amplitude_contrast=amp,
            voltage=vol,
            spherical_aberration=cs,
            phase_shift_deg=0.0,
        )
    )

ACCUMULATED_DOSE = read_dose_file(
    pathlib.Path(__file__).parent.joinpath("Data").joinpath("test_dose.txt")
)
TILT_ANGLES = read_tlt_file(
    pathlib.Path(__file__).parent.joinpath("Data").joinpath("test_angles.rawtlt")
)


def make_relion5_tomo_stars(
    tomoname: str,
    out_dir: pathlib.Path,
    voltage: int = vol,
    amp_contrast: float = amp,
    cs: float = cs,
    tomohand: int = 1,
    pixelsize: float = 1.0,
    binning: float = 1.0,
    tilt_angles: list = TILT_ANGLES,
    dose: list = ACCUMULATED_DOSE,
    defocus: list = defocus_data,
):
    """This is a function that outputs a tomogram.star and a tilt_series/tilt_serie.star
    Mostly to test our relion5 processing is (internally) consistent

    Parameters:

    """
    if not out_dir.is_dir():
        out_dir.mkdir(parents=True)

    tomo_star = out_dir / "tomogram.star"
    tilt_star = out_dir / "tilt_series" / "tilt_serie.star"
    tilt_star.parent.mkdir()

    tilt_data = [
        [tilt_angle, do, de, de]
        for do, tilt_angle, de in zip(dose, tilt_angles, defocus)
    ]
    tilt_output = pd.DataFrame(
        tilt_data,
        columns=[
            "rlnTomoNominalStageTiltAngle",
            "rlnMicrographPreExposure",
            "rlnDefocusV",
            "rlnDefocusU",
        ],
    )
    tomo_data = [
        [
            tomoname,
            voltage,
            cs,
            amp_contrast,
            tomohand,
            pixelsize,
            str(tilt_star),
            binning,
        ]
    ]
    tomo_output = pd.DataFrame(
        tomo_data,
        columns=[
            "rlnTomoName",
            "rlnVoltage",
            "rlnSphericalAberration",
            "rlnAmplitudeContrast",
            "rlnTomoHand",
            "rlnTomoTiltSeriesPixelSize",
            "rlnTomoTiltSeriesStarFile",
            "rlnTomoTomogramBinning",
        ],
    )
    starfile.write({"global": tomo_output}, tomo_star, overwrite=True)
    starfile.write({tomoname: tilt_output}, tilt_star, overwrite=True)


def make_random_particles(n: int = 10, relion5: bool = False) -> pd.DataFrame:
    """This is a function that outputs completely random data for testing purposses

    Parameters
    ----------
    n: int
        number of particles to generate random data for
    relion5: bool, default False
        output a relion5 compatible dataframe instead of a relion4 one (default)

    Returns
    -------
    dataframe: pd.DataFrame
        dataframe with data that can be written out as a STAR file
    """
    # Some common variables
    cut_off, sigma = np.random.rand(2)
    pixel_size = 0.5 + 2 * np.random.rand()  # between 0.5 and 2.5
    tomogram_id = f"Tomo{np.random.randint(2147483647)}"  # np.int32 max
    common = [cut_off, sigma, pixel_size, tomogram_id]
    data = []
    for i in range(n):
        x, y, z = np.random.randint(128, size=3)
        rot, psi = 2 * np.pi * np.random.rand(2)
        tilt = np.pi * np.random.rand()
        lcc_max = cut_off + (1 - cut_off) * np.random.rand()
        temp = [x, y, z, rot, tilt, psi, lcc_max] + common
        data.append(temp)

    output = pd.DataFrame(
        data,
        columns=[
            "rlnCoordinateX",
            "rlnCoordinateY",
            "rlnCoordinateZ",
            "rlnAngleRot",
            "rlnAngleTilt",
            "rlnAnglePsi",
            "rlnLCCmax",
            "rlnCutOff",
            "rlnSearchStd",
            "rlnDetectorPixelSize",
            "rlnMicrographName",
        ],
    )
    if relion5:
        # Use same alterations as in extract.py
        center = 128 / 2
        output["rlnCoordinateX"], output["rlnCoordinateY"], output["rlnCoordinateZ"] = (
            (output["rlnCoordinateX"] - center) * pixel_size,
            (output["rlnCoordinateY"] - center) * pixel_size,
            (output["rlnCoordinateZ"] - center) * pixel_size,
        )
        column_change = {
            "rlnCoordinateX": "rlnCenteredCoordinateXAngst",
            "rlnCoordinateY": "rlnCenteredCoordinateYAngst",
            "rlnCoordinateZ": "rlnCenteredCoordinateZAngst",
            "rlnMicrographName": "rlnTomoName",
            "rlnDetectorPixelSize": "rlnTomoTiltSeriesPixelSize",
        }
        output = output.rename(columns=column_change)
    return output


@contextlib.contextmanager
def chdir(directory):
    """This is a function that allows a test to use chdir,
    but not error tests run after it if an error is raised while chdir is active

    Parameters
    ----------
    directory: pathlib.Path
        directory to chdir into

    Returns
    -------
    contextmanager: contextlib.contextmanager
        a contextmanager that deals with unwinding chdir if an
        error is raised while the context is active
    """
    old = os.getcwd()
    os.chdir(directory)
    try:
        yield
    finally:
        os.chdir(old)



import unittest
import pathlib
import starfile
import glob
from tempfile import TemporaryDirectory
import pandas as pd
from pytom_tm.entry_points import merge_stars
from pytom_tm.utils import mute_stdout_stderr
from testing_utils import make_random_particles, chdir


class TestMergeStars(unittest.TestCase):
    def setUp(self):
        tempdir = TemporaryDirectory()
        self.addCleanup(tempdir.cleanup)
        self.dirname = tempdir.name
        self.tempdir = pathlib.Path(tempdir.name)
        self.n = 10
        particles1 = make_random_particles(n=self.n)
        particles2 = make_random_particles(n=self.n)
        self.particles = [particles1, particles2]

    def test_error_function_on_empty(self):
        from pytom_tm.merge_stars import merge_stars as ms

        # make sure we raise if we don't find any starfiles
        with self.assertRaisesRegex(ValueError, "No starfiles"):
            ms([], self.tempdir / "not_relevant.star")

    def test_relion4_mode(self):
        # write 2 starfiles
        for particle in self.particles:
            tomo_id = particle["rlnMicrographName"][0]
            starfile.write(
                {"particles": particle}, self.tempdir / f"{tomo_id}_particles.star"
            )

        outfile = str(self.tempdir / "test.star")
        # Make a joined file via the entry point
        # mimick star expansion on a bash shell
        in_files = glob.glob(f"{self.dirname}/*.star")
        merge_stars(["-i"] + in_files + ["-o", outfile])

        # make sure we can read the output starfile
        out = starfile.read(outfile)

        # make sure we have the number of expected lines
        self.assertEqual(self.n * 2, len(out))

        # check that both tomo IDs are in the new starfile
        for particle in self.particles:
            for name in set(particle["rlnMicrographName"]):
                self.assertIn(name, set(out["rlnMicrographName"]))

        # test that we fail if we just give one starfile in this mode
        with self.assertRaisesRegex(ValueError, "doesn't make sense to merge"):
            merge_stars(["-i", in_files[0], "-o", outfile])

    def test_multi_datablock_file(self):
        # write 2 starfiles
        for i, particle in enumerate(self.particles):
            tomo_id = particle["rlnMicrographName"][0]
            # make a second dataframe
            test_data = pd.DataFrame(
                [[f"test{i}", tomo_id]], columns=["testID", "TomoID"]
            )
            starfile.write(
                {"testdata": test_data, "particles": particle},
                self.tempdir / f"{tomo_id}_particles.star",
            )

        outfile = str(self.tempdir / "test.star")
        # Make a joined file via the entry point
        # mimick star expansion on a bash shell
        in_files = glob.glob(f"{self.dirname}/*.star")
        with self.assertLogs(level="WARNING") as cm:
            merge_stars(["-i"] + in_files + ["-o", outfile, "--log-test"])
        for o in cm.output:
            if "multi-data-block" in o:
                break
        else:
            self.fail("expected a warning about multi-data-block")  # pragma: no cover
        # make sure we can read the output starfile
        out = starfile.read(outfile)
        self.assertIs(type(out), pd.DataFrame)
        # make sure we have the number of expected lines
        self.assertEqual(self.n * 2, len(out))

        # check that both tomo IDs are in the new starfile
        for particle in self.particles:
            for name in set(particle["rlnMicrographName"]):
                self.assertIn(name, set(out["rlnMicrographName"]))

        # redo for relion5
        with TemporaryDirectory() as outdir:
            outdir_path = pathlib.Path(outdir)
            particles1 = make_random_particles(relion5=True)
            particles2 = make_random_particles(relion5=True)
            for i, particle in enumerate([particles1, particles2]):
                tomo_id = particle["rlnTomoName"][0]
                test_data = pd.DataFrame(
                    [[f"test{i}", tomo_id]], columns=["testID", "TomoID"]
                )
                starfile.write(
                    {"testdata": test_data, "particles": particle},
                    outdir_path / f"{tomo_id}_particles.star",
                )

            outfile = str(outdir_path / "test.star")

            in_files = glob.glob(f"{outdir}/*.star")

            # Make a joined file via the entry point
            with self.assertNoLogs(level="WARNING"):
                merge_stars(
                    ["-i"]
                    + in_files
                    + ["-o", outfile, "--relion5-compat", "--log-test"]
                )

            # make sure we can read the output starfile
            out = starfile.read(outfile)
            # make sure that the outfile is as expected:
            # - 2 columns with names _rlnTomoName and _rlnTomoImportParticleFile
            self.assertIs(type(out), pd.DataFrame)
            self.assertEqual(2, len(out.columns))
            for col in ["rlnTomoName", "rlnTomoImportParticleFile"]:
                self.assertIn(col, out.columns)

            # check that both tomo IDs are in the new starfile and the correct filename
            for particle in [particles1, particles2]:
                for name in set(particle["rlnTomoName"]):
                    self.assertIn(name, set(out["rlnTomoName"]))

        # make sure we fail on no particles
        with TemporaryDirectory() as outdir:
            outdir_path = pathlib.Path(outdir)
            particles1 = make_random_particles(relion5=True)
            particles2 = make_random_particles(relion5=True)
            for i, particle in enumerate([particles1, particles2]):
                tomo_id = particle["rlnTomoName"][0]
                test_data = pd.DataFrame(
                    [[f"test{i}", tomo_id]], columns=["testID", "TomoID"]
                )
                starfile.write(
                    {"testdata": test_data, "failparticles": particle},
                    outdir_path / f"{tomo_id}_particles.star",
                )

            outfile = str(outdir_path / "test.star")

            in_files = glob.glob(f"{outdir}/*.star")

            # Make sure we fail joining the file via the entry point
            with self.assertRaisesRegex(ValueError, "'particles' data block"):
                merge_stars(["-i"] + in_files + ["-o", outfile, "--relion5-compat"])

    def test_fail_on_incompatible_starfiles(self):
        # Make sure we fail if we try to combine RELION4 starfiles
        # with a RELION5 flag

        # write 2 relion 4 starfiles
        for particle in self.particles:
            tomo_id = particle["rlnMicrographName"][0]
            starfile.write(
                {"particles": particle}, self.tempdir / f"{tomo_id}_particles.star"
            )

        outfile = str(self.tempdir / "test.star")
        # Make sure entry point fails if trying to do relion5 merge
        in_files = glob.glob(f"{self.dirname}/*.star")
        with self.assertRaisesRegex(ValueError, "rlnTomoName"):
            merge_stars(["-i"] + in_files + ["-o", outfile, "--relion5-compat"])

    def test_relion5_mode(self):
        particles1 = make_random_particles(relion5=True)
        particles2 = make_random_particles(relion5=True)
        for particle in [particles1, particles2]:
            tomo_id = particle["rlnTomoName"][0]
            starfile.write(
                {"particles": particle}, self.tempdir / f"{tomo_id}_particles.star"
            )

        outfile = str(self.tempdir / "test.star")
        # Make sure entry point fails if trying to do relion5 merge
        in_files = glob.glob(f"{self.dirname}/*.star")

        # Make a joined file via the entry point
        merge_stars(["-i"] + in_files + ["-o", outfile, "--relion5-compat"])

        # make sure we can read the output starfile
        out = starfile.read(outfile)
        # make sure that the outfile is as expected:
        # - 2 columns with names _rlnTomoName and _rlnTomoImportParticleFile
        self.assertEqual(2, len(out.columns))
        for col in ["rlnTomoName", "rlnTomoImportParticleFile"]:
            self.assertIn(col, out.columns)

        # check that both tomo IDs are in the new starfile and the correct filename
        for particle in [particles1, particles2]:
            for name in set(particle["rlnTomoName"]):
                self.assertIn(name, set(out["rlnTomoName"]))

        # Check that all star files can be read and contains the expected tomoname
        for _, (tomoname, filename) in out.iterrows():
            temp = starfile.read(filename)
            temp_tomoname = set(temp["rlnTomoName"])
            self.assertIn(tomoname, temp_tomoname)

        # Recheck that we can read all the star files even if we are in a different cwd
        tempdir2 = TemporaryDirectory()
        with chdir(tempdir2.name):
            for _, (tomoname, filename) in out.iterrows():
                temp = starfile.read(filename)
                temp_tomoname = set(temp["rlnTomoName"])
                self.assertIn(tomoname, temp_tomoname)
        tempdir2.cleanup()

        # test that we pass if we just give one starfile in this mode
        outfile2 = str(self.tempdir / "single_test.star")
        merge_stars(["-i", in_files[0], "-o", outfile2, "--relion5-compat"])

        # make sure we only written a single line
        out2 = starfile.read(outfile2)
        self.assertEqual(1, len(out2))

    def test_fail_on_dir_input(self):
        with mute_stdout_stderr(), self.assertRaises(SystemExit) as ex:
            merge_stars(["-i", f"{self.dirname}", "-o", "irrelevant.star"])
        self.assertEqual(ex.exception.code, 2)

    def test_multi_dir_input(self):
        # write 2 starfiles
        directories = [self.tempdir / f"test{i}" for i in range(2)]
        for particle, directory in zip(self.particles, directories):
            directory.mkdir()
            tomo_id = particle["rlnMicrographName"][0]
            starfile.write(
                {"particles": particle}, directory / f"{tomo_id}_particles.star"
            )

        outfile = str(self.tempdir / "test.star")

        # Mimick inputting `output_dir/test*/*.star`
        star_files = glob.glob(f"{self.dirname}/test*/*.star")

        merge_stars(["-i"] + star_files + ["-o", outfile])

        # make sure we can read the output starfile
        out = starfile.read(outfile)

        # make sure we have the number of particles we expect
        self.assertEqual(len(out), 2 * self.n)

    def test_non_unique_input(self):
        # write 2 starfiles
        for particle in self.particles:
            tomo_id = particle["rlnMicrographName"][0]
            starfile.write(
                {"particles": particle}, self.tempdir / f"{tomo_id}_particles.star"
            )

        outfile = str(self.tempdir / "test.star")
        # Make a joined file via the entry point
        # mimick star expansion on a bash shell
        in_files = glob.glob(f"{self.dirname}/*.star")
        with self.assertLogs(level="WARNING") as cm:
            # Give all the files twice
            merge_stars(["-i"] + in_files + in_files + ["-o", outfile, "--log-test"])
        for o in cm.output:
            if "duplicate input" in o:
                break
            else:
                self.fail("expected warning is not logged")  # pragma: no cover

        # make sure we can read the output starfile
        out = starfile.read(outfile)

        # make sure we only have the number of expected lines from the unique files
        self.assertEqual(self.n * 2, len(out))



import unittest
import pathlib
from pytom_tm.angles import load_angle_list, angle_to_angle_list
import numpy as np
import itertools as itt
import re
import logging

TEST_DATA_DIR = pathlib.Path(__file__).parent.joinpath("test_data")
ERRONEOUS_ANGLE_FILE = TEST_DATA_DIR.joinpath("error_angles.txt")
UNORDERED_ANGLE_FILE = TEST_DATA_DIR.joinpath("unordered_angles.txt")


class TestAngles(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        TEST_DATA_DIR.mkdir()
        # create an erroneous angle file
        with open(ERRONEOUS_ANGLE_FILE, "w") as fstream:
            fstream.write(" ".join(map(str, [1.0] * 4)) + "\n")
            fstream.write(" ".join(map(str, [1.0] * 3)) + "\n")
        # create an unordered angle file
        with open(UNORDERED_ANGLE_FILE, "w") as fstream:
            fstream.write(" ".join(["3.", "3.", "1."]) + "\n")
            fstream.write(" ".join(["3", "2.", "1."]) + "\n")
            fstream.write(" ".join(["2.", "3.", "1."]) + "\n")
            fstream.write(" ".join(["3.", "2.", "2."]) + "\n")

    @classmethod
    def tearDownClass(cls) -> None:
        for f in [ERRONEOUS_ANGLE_FILE, UNORDERED_ANGLE_FILE]:
            f.unlink()
        TEST_DATA_DIR.rmdir()

    def test_load_list(self):
        with self.assertRaisesRegex(
            ValueError,
            "each line should have 3",
            msg="Invalid angle file should raise an error",
        ):
            load_angle_list(ERRONEOUS_ANGLE_FILE)

    def test_load_sort(self):
        angles = load_angle_list(UNORDERED_ANGLE_FILE, sort_angles=True)
        expected = [(2.0, 3.0, 1.0), (3.0, 2.0, 1.0), (3.0, 2.0, 2.0), (3.0, 3.0, 1.0)]
        self.assertEqual(angles, expected)

    def test_angle_to_angle_list(self):
        # ask for a random sample between [1 - 90)
        angle = 1 + np.random.random() * 89
        with self.assertLogs(level="INFO") as cm:
            angles = angle_to_angle_list(angle, log_level=logging.INFO)

        # Check logs and if all angles are smaller or equal
        self.assertEqual(len(cm.output), 2)
        for out in cm.output:
            # Check if the used_angle <= input angle
            # regex = whitespace -> 0-inf digits -> "." -> 1-inf digits -> whitespace
            possible_match = re.findall(r"\s\d*[.]\d+\s", out)
            self.assertEqual(len(possible_match), 1)
            self.assertLessEqual(float(possible_match[0]), angle)

        # make sure everything is sorted and X is never 0
        for a, b in itt.pairwise(angles):
            # make sure default is sorted
            self.assertLess(a, b)
            # make sure X is never 0
            self.assertNotEqual(a[1], 0)
        # also check the last X
        self.assertNotEqual(b[1], 0)



import unittest
import voltools as vt
import cupy as cp
from pytom_tm.mask import spherical_mask, ellipsoidal_mask
from pytom_tm.angles import angle_to_angle_list
from pytom_tm.correlation import normalised_cross_correlation


class TestMask(unittest.TestCase):
    def setUp(self):
        self.angles = angle_to_angle_list(50.00)

    def test_rotational_invariance_even(self):
        # TEST EVEN MASK
        nxcc_offcenter, nxcc_centered = [], []

        mask = cp.asarray(spherical_mask(12, 4, 0.5))
        mask_rotated = cp.zeros_like(mask)

        mask_texture = vt.StaticVolume(mask, interpolation="filt_bspline", device="gpu")

        for i in range(len(self.angles)):
            mask_texture.transform(
                rotation=self.angles[i],
                rotation_units="rad",
                rotation_order="rzxz",
                center=tuple([x // 2 for x in mask.shape]),
                output=mask_rotated,
            )

            nxcc_offcenter.append(
                normalised_cross_correlation(mask, mask_rotated).get()
            )

        for i in range(len(self.angles)):
            mask_texture.transform(
                rotation=self.angles[i],
                rotation_units="rad",
                rotation_order="rzxz",
                output=mask_rotated,
                # center=np.divide(np.subtract(mask.shape, 1), 2, dtype=np.float32),
            )

            nxcc_centered.append(normalised_cross_correlation(mask, mask_rotated).get())

        self.assertTrue(
            sum(nxcc_centered) > sum(nxcc_offcenter),
            msg="Center of rotation for mask is incorrect.",
        )
        self.assertTrue(
            sum(nxcc_centered) > 99.27, msg="Precision of mask rotation is too low."
        )

    def test_rotational_invariance_uneven(self):
        # TEST UNEVEN MASK
        nxcc_offcenter, nxcc_centered = [], []

        mask = cp.asarray(spherical_mask(13, 4, 0.5))
        mask_rotated = cp.zeros_like(mask)

        mask_texture = vt.StaticVolume(mask, interpolation="filt_bspline", device="gpu")

        for i in range(len(self.angles)):
            mask_texture.transform(
                rotation=self.angles[i],
                rotation_units="rad",
                rotation_order="rzxz",
                center=tuple([x // 2 for x in mask.shape]),
                output=mask_rotated,
            )

            nxcc_offcenter.append(
                normalised_cross_correlation(mask, mask_rotated).get()
            )

        for i in range(len(self.angles)):
            mask_texture.transform(
                rotation=self.angles[i],
                rotation_units="rad",
                rotation_order="rzxz",
                output=mask_rotated,
                # center=np.divide(np.subtract(mask.shape, 1), 2, dtype=np.float32),
            )

            nxcc_centered.append(normalised_cross_correlation(mask, mask_rotated).get())

        self.assertAlmostEqual(
            sum(nxcc_centered),
            sum(nxcc_offcenter),
            places=4,
            msg="Center of rotation for mask is incorrect.",
        )
        self.assertTrue(
            sum(nxcc_centered) > 99.09, msg="Precision of mask rotation is too low."
        )

    def test_ellipsoidal_mask_errors(self):
        # Make sure we error on negative box_size, major, minor1, or minor2
        default = {
            "box_size": 16,
            "major": 12,
            "minor1": 8,
            "minor2": 6,
            "smooth": 3,
            "cutoff_sd": 2,
        }
        for i in ["box_size", "major", "minor1", "minor2"]:
            inp = default.copy()
            inp[i] *= -1
            with self.assertRaisesRegex(ValueError, "box_size or radii"):
                ellipsoidal_mask(**inp)
        # Make sure we error on negative smooth or cutoff_sd
        for i in ["smooth", "cutoff_sd"]:
            inp = default.copy()
            inp[i] *= -1
            with self.assertRaisesRegex(ValueError, "smooth or sd cutoff"):
                ellipsoidal_mask(**inp)



import unittest
import json
from pytom_tm.json import JsonSerializable, CustomJSONEncoder


class TestJSON(unittest.TestCase):
    def test_no_class(self):
        test_dict = {"a": 123}
        with self.assertRaisesRegex(ValueError, "None"):
            _ = JsonSerializable._rebuild_from_dict(test_dict)

    def test_custom_encoder(self):
        class Test1:
            def __init__(self):
                self.test = 12

        class Test2(JsonSerializable):
            def __init__(self):
                self.test = 42

        # Make sure the Encoder raises as expected
        test1 = Test1()
        with self.assertRaisesRegex(TypeError, "Test1"):
            _ = json.dumps(test1, cls=CustomJSONEncoder)

        test2 = Test2()
        out = json.dumps(test2, cls=CustomJSONEncoder)
        self.assertIn("__class__", out)
        self.assertIn("Test2", out)
        self.assertIn(Test2, JsonSerializable._registry.values())
        self.assertIn("42", out)



import json
from typing import Any, Type


class JsonSerializable:
    """Mixin class to make a class JSON serializable together with the custom Encoder
    and Decoder defined in this file.

    It uses __init_subclass__, which is called whenever a subclass is
    defined, so we can pre-register which classes we expect to load in this way
    """

    _registry: dict[str, Type["JsonSerializable"]] = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        JsonSerializable._registry[cls.__name__] = cls

    @classmethod
    def _rebuild_from_dict(cls, data: dict[str, Any]) -> "JsonSerializable":
        """Rebuild any registered class from a dict."""
        class_name = data.pop("__class__", None)
        if class_name not in cls._registry:
            raise ValueError(f"Unknown class: {class_name}")
        target_cls = cls._registry[class_name]
        return target_cls(**data)


class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, JsonSerializable):
            result = obj.__dict__.copy()
            result["__class__"] = obj.__class__.__name__
            return result
        return super().default(obj)


class CustomJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        super().__init__(object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        if "__class__" in obj:
            return JsonSerializable._rebuild_from_dict(obj)
        return obj
